# pylint: disable = import-outside-toplevel
"""Tests of ``scripts.cioupdate`` function and class."""

from os.path import exists, basename, join
from collections import namedtuple
from warnings import catch_warnings, simplefilter
from getpass import getuser
from unittest import TestCase

from testfixtures import log_capture

from . import TmpDirTestCase


# =============================================================================
class UScriptsCioUpdateMain(TestCase):
    """Unit test class for :func:`scripts.cioupdate.main`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:scripts.cioupdate.main]"""
        from ..scripts.cioupdate import main
        from . import CIOUPDATE_INI

        self.assertRaises(SystemExit, main, [])
        main(['foo.ini'])
        main([CIOUPDATE_INI])


# =============================================================================
class UScriptsCioUpdateCioUpdate(TmpDirTestCase):
    """Unit test class for :class:`scripts.cioupdate.CioUpdate`."""

    _currentdir = None

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        super(UScriptsCioUpdateCioUpdate, self).setUp()

        self.args = namedtuple(
            'ArgumentParser',
            'lang password no_backup no_update log_level log_file'
            ' drop_tables remove_locks remove_builds skip_refresh'
            ' recreate_thumbnails reindex extra key')(
                lang='en', password=None, no_backup=False, no_update=False,
                log_level='INFO', log_file=None, drop_tables=False,
                remove_locks=False, remove_builds=False, skip_refresh=False,
                recreate_thumbnails=False, reindex=False, extra=(), key=None)

    # -------------------------------------------------------------------------
    def test_start_encrypt(self):
        """[u:scripts.cioupdate.CioUpdate.start] encrypt"""
        from . import CIOUPDATE_INI
        from ..scripts.cioupdate import CioUpdate

        @log_capture()
        def call_test(log):
            """Function to ignore DeprecationWarning."""
            args = namedtuple('ArgumentParser', 'conf_file password key')(
                conf_file=CIOUPDATE_INI, password='foo', key=None)
            cioupdate = CioUpdate(args, CIOUPDATE_INI)
            self.assertTrue(cioupdate.start())
            self.assertEqual(len(log.records), 1)
            self.assertIn('foo = ', log.records[-1].getMessage())

        with catch_warnings():
            simplefilter('ignore')
            # pylint: disable = no-value-for-parameter
            call_test()

    # -------------------------------------------------------------------------
    def test_start_ok(self):
        """[u:scripts.cioupdate.CioUpdate.start] Ok"""
        from . import TEST_DIR, CIOUPDATE_INI, BACKUP_DIR
        from ..scripts.cioupdate import CioUpdate

        self._make_database()
        with open(CIOUPDATE_INI, 'r') as hdl:
            content = hdl.read()
        content = content\
            .replace('%(here)s/../..', '%(here)s/..')\
            .replace('#user =', 'user = {0}'.format(getuser()))
        cioupdate_ini = join(TEST_DIR, basename(CIOUPDATE_INI))
        with open(cioupdate_ini, 'w') as hdl:
            hdl.write(content)

        @log_capture()
        def call_test(log):
            """Function to ignore DeprecationWarning."""
            cioupdate = CioUpdate(self.args, cioupdate_ini)
            cioupdate.start()
            messages = '\n'.join([k.getMessage() for k in log.records])
            self.assertIn('ciobackup TestChrysalio', messages)
            self.assertIn('/Chrysalio', messages)
            print(messages)
            self.assertTrue(
                ('ciopopulate TestChrysalio' in messages) or
                ('Temporary failure' in messages) or
                ('uthorization failed' in messages) or
                ('Connection timed out' in messages) or
                ('No route to host') in messages or
                ('Could not resolve host') in messages or
                ('cannot pull with rebase') in messages)
            self.assertTrue(exists(BACKUP_DIR))

        with catch_warnings():
            simplefilter('ignore')
            # pylint: disable = no-value-for-parameter
            call_test()

    # -------------------------------------------------------------------------
    def test_start_invalid_backup_user(self):
        """[u:scripts.cioupdate.CioUpdate.start] Invalid backup user"""
        from . import CIOUPDATE1_INI
        from ..scripts.cioupdate import CioUpdate

        self._make_database()

        @log_capture()
        def call_test(log):
            """Function to ignore DeprecationWarning."""
            cioupdate = CioUpdate(self.args, CIOUPDATE1_INI)
            cioupdate.start(['TestChrysalio1'])
            self.assertFalse([
                True for k in log.records if k.levelname == 'ERROR'])
            cioupdate.start(['TestChrysalio2'])
            self.assertTrue([
                True for k in log.records if k.levelname == 'ERROR'])

        with catch_warnings():
            simplefilter('ignore')
            # pylint: disable = no-value-for-parameter
            call_test()

    # -------------------------------------------------------------------------
    def test_start_invalid_update_user(self):
        """[u:scripts.cioupdate.CioUpdate.start] Invalid update user"""
        from . import CIOUPDATE2_INI
        from ..scripts.cioupdate import CioUpdate

        @log_capture()
        def call_test(log):
            """Function to ignore DeprecationWarning."""
            cioupdate = CioUpdate(self.args, CIOUPDATE2_INI)
            self.assertFalse(cioupdate.start())
            messages = '\n'.join([k.getMessage() for k in log.records])
            self.assertTrue(
                ('name not found' in messages) or
                ('Temporary failure' in messages) or
                ('uthorization failed' in messages) or
                ('Connection timed out' in messages) or
                ('No route to host') in messages or
                ('Could not resolve host') in messages or
                ('cannot pull with rebase') in messages)

        with catch_warnings():
            simplefilter('ignore')
            # pylint: disable = no-value-for-parameter
            call_test()

    # -------------------------------------------------------------------------
    def test_start_incorrect_repository(self):
        """[u:scripts.cioupdate.CioUpdate.start] Incorrect repository"""
        from . import CIOUPDATE3_INI
        from ..scripts.cioupdate import CioUpdate

        @log_capture()
        def call_test(log):
            """Function to ignore DeprecationWarning."""
            cioupdate = CioUpdate(self.args, CIOUPDATE3_INI)
            self.assertFalse(cioupdate.start())
            self.assertIn(
                'Git', '\n'.join([k.getMessage() for k in log.records]))

        with catch_warnings():
            simplefilter('ignore')
            # pylint: disable = no-value-for-parameter
            call_test()

    # -------------------------------------------------------------------------
    def test_start_invalid_populate(self):
        """[u:scripts.cioupdate.CioUpdate.start] Invalid populate command"""
        from . import CIOUPDATE4_INI
        from ..scripts.cioupdate import CioUpdate

        @log_capture()
        def call_test(log):
            """Function to ignore DeprecationWarning."""
            cioupdate = CioUpdate(self.args, CIOUPDATE4_INI)
            self.assertFalse(cioupdate.start())
            self.assertIn(
                'foopopulate', '\n'.join(
                    [k.getMessage() for k in log.records]))

        with catch_warnings():
            simplefilter('ignore')
            # pylint: disable = no-value-for-parameter
            call_test()

    # -------------------------------------------------------------------------
    def test_start_invalid_backup(self):
        """[u:scripts.cioupdate.CioUpdate.start] Invalid backup command"""
        from . import CIOUPDATE5_INI
        from ..scripts.cioupdate import CioUpdate

        @log_capture()
        def call_test(log):
            """Function to ignore DeprecationWarning."""
            cioupdate = CioUpdate(self.args, CIOUPDATE5_INI)
            self.assertFalse(cioupdate.start())
            self.assertIn(
                'foobackup', '\n'.join([k.getMessage() for k in log.records]))

        with catch_warnings():
            simplefilter('ignore')
            # pylint: disable = no-value-for-parameter
            call_test()

    # -------------------------------------------------------------------------
    def test_start_empty_command(self):
        """[u:scripts.cioupdate.CioUpdate.start] Empty command"""
        from . import CIOUPDATE6_INI
        from ..scripts.cioupdate import CioUpdate

        cioupdate = CioUpdate(self.args, CIOUPDATE6_INI)
        self.assertTrue(cioupdate.start())

    # -------------------------------------------------------------------------
    @classmethod
    def _make_database(cls):
        """Create a temporary database."""
        from . import CIOUPDATE_INI
        from ..relaxng import RELAXNG
        from ..scripts.ciopopulate import Populate
        from ..models.dbuser import DBUser
        from ..models.populate import xml2db

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang attachments')(
                conf_uri=CIOUPDATE_INI, options=None, lang=None,
                attachments=None)
        Populate(args, DBUser, xml2db, RELAXNG).run()



"""An exemple of a module with unsatisfied dependency."""

from ...modules import Module


# =============================================================================
def includeme(configurator):
    """Function to register the module.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    # Registration
    Module.register(configurator, BadDependencyModule)


# =============================================================================
class BadDependencyModule(Module):
    """Class for Chrysalio module."""

    name = 'Bad dependencies'
    dependencies = ('chrysalio.modules.foo',)



"""CioSkeleton mode."""

from ...lib.i18n import _


# Mode = (mode_id, (icon, label, permission, route, menu))
MODE_CIOSKELETON = ('skeleton', (
    '{theme}/cioskeleton/images/menu_cioskeleton.png', _('Skeleton'),
    'mode-skeleton', 'bone_index', None))



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioskeleton.models.populate`` functions."""

from ....tests import DBUnitTestCase


# =============================================================================
class UModelsPopulateXml2Db(DBUnitTestCase):
    """Unit test class for
    :func:`modules.cioskeleton.models.populate.xml2db`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:modules.cioskeleton.models.populate.xml2db]"""
        from lxml import etree
        from ..models.populate import xml2db

        tree = etree.XML(
            '<cioskeleton'
            '    xmlns="http://ns.chrysal.io/cioskeleton" version="1.0">'
            '  <bones>'
            '    <bone id="hand">'
            '      <label>Hand</label>'
            '      <foo/>'
            '    </bone>'
            '  </bones>'
            '</cioskeleton>')
        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member

        self.assertEqual(len(xml2db(dbsession, tree, None, True)), 0)


# =============================================================================
class UModelsPopulateDb2Xml(DBUnitTestCase):
    """Unit test class for
    :func:`modules.cioskeleton.models.populate.db2xml`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:modules.cioskeleton.models.populate.db2xml]"""
        from lxml import etree
        from ..models.populate import db2xml
        from ..models.dbbone import DBBone

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        dbsession.add(DBBone(bone_id='hand', label='Hand'))
        namespace = 'http://ns.chrysal.io/cioskeleton'
        root_elt = etree.Element(
            '{{{0}}}cioskeleton'.format(namespace),
            nsmap={None: namespace}, version='1.0')

        db2xml(dbsession, root_elt)
        self.assertEqual(len(root_elt), 1)



# pylint: disable = import-outside-toplevel
"""Tests of ``includes.modules.models.DBModule`` class."""

from ....tests import DBUnitTestCase


# =============================================================================
class UIncludesModulesModelsDBModule(DBUnitTestCase):
    """Unit test class for :class:`includes.modules.models.DBModule`."""

    # -------------------------------------------------------------------------
    def test_table4view(self):
        """[u:includes.modules.models.DBModule.table4view]"""
        from collections import OrderedDict
        from ....tests import TEST_INI
        from ....modules.cioldap import ModuleCioLDAP
        from ..models import DBModule

        html = DBModule.table4view(self.request)
        self.assertEqual(html, 'No module.')

        self.configurator.add_route('cioldap_view', '/cioldap/view')
        self.request.registry['modules'] = OrderedDict((
            ('chrysalio.modules.cioldap', ModuleCioLDAP(TEST_INI)),))
        self.request.registry['modules_off'] = set()

        html = DBModule.table4view(self.request)
        self.assertIn('title="chrysalio.modules.cioldap"', html)
        self.assertIn('>LDAP</', html)

    # -------------------------------------------------------------------------
    def test_settings_schema(self):
        """[u:includes.modules.models.DBModule.settings_schema]"""
        from collections import OrderedDict
        from colander import SchemaNode
        from ....tests import TEST_INI
        from ....modules.cioskeleton import ModuleCioSkeleton
        from ..models import DBModule

        self.request.registry['modules'] = OrderedDict((
            ('chrysalio.modules.cioskeleton', ModuleCioSkeleton(TEST_INI)),))
        self.request.registry['modules_off'] = set()

        schema, defaults = DBModule.settings_schema(self.request)
        self.assertIsInstance(schema, SchemaNode)
        self.assertIn('chrysalio.modules.cioskeleton', schema.serialize())
        self.assertIsInstance(defaults, dict)
        self.assertIn('chrysalio.modules.cioskeleton', defaults)
        self.assertTrue(defaults['chrysalio.modules.cioskeleton'])

    # -------------------------------------------------------------------------
    def test_table4edit(self):
        """[u:includes.modules.models.DBModule.table4edit]"""
        from collections import OrderedDict
        from ....tests import TEST_INI
        from ....lib.form import Form
        from ....modules.cioskeleton import ModuleCioSkeleton
        from ..models import DBModule

        html = DBModule.table4edit(self.request, Form(self.request))
        self.assertEqual(html, 'No module.')

        self.request.registry['modules'] = OrderedDict((
            ('chrysalio.modules.cioskeleton', ModuleCioSkeleton(TEST_INI)),))
        self.request.registry['modules_off'] = set(
            ['chrysalio.modules.cioskeleton'])

        html = DBModule.table4edit(self.request, Form(self.request))
        self.assertIn('chrysalio.modules.cioskeleton', html)
        self.assertIn('>Skeleton</', html)



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioldap.models.dbldap`` classes."""

from ....tests import DBUnitTestCase
from ..models.dbldap import DBLdap


# =============================================================================
class UModelsDBLdapDBLdap(DBUnitTestCase):
    """Unit test class for :class:`modules.cioldap.models.dbldap.DBLdap`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _make_one(cls):
        """Make an `DBLdap`` object."""
        from ..models.dbldap import DBLdapProfile

        dbldap = DBLdap(
            ssl=True,
            base='dc=iinov,dc=com',
            root_dn='cn=admin,dc=iinov,dc=com',
            root_password='/ZSHiyU30mLjsr20KSXUoAD3tcragNnsQY+YxS2cIiA=',
            user_dn='uid=_UID_,ou=people,dc=iinov,dc=com',
            user_filter='(&(objectclass=inetOrgPerson)(uid=_UID_))',
            field_first_name='givenName',
            field_last_name='sn',
            field_email='mail')
        dbldap.user_profiles.append(DBLdapProfile(profile_id='profile_viewer'))

        return dbldap

    # -------------------------------------------------------------------------
    def test_xml2db(self):
        """[u:.modules.cioldap.models.dbldap.DBLdap.xml2db]"""
        from lxml import etree

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        dbsession.add(DBLdap(
            base='dc=iinov,dc=fr',
            user_dn='uid=_UID_,ou=people,dc=iinov,dc=com',
            user_filter='(&(objectclass=inetOrgPerson)(uid=_UID_))',
            field_first_name='givenName'))

        # Without element
        self.assertIsNone(DBLdap.xml2db(dbsession, None))
        self.assertTrue(dbsession.query(DBLdap).first())

        # With an incorrect configuration
        ldap_elt = etree.XML(
            '<ldap xmlns="http://ns.chrysal.io/cioldap">'
            '  <user>'
            '    <dn>uid=_UID_,ou=people,dc=iinov,dc=com</dn>'
            '  </user>'
            '</ldap>')
        self.assertIsNotNone(DBLdap.xml2db(dbsession, ldap_elt))
        self.assertFalse(dbsession.query(DBLdap).first())

        # With a correct configuration
        ldap_elt = etree.XML(
            '<ldap xmlns="http://ns.chrysal.io/cioldap">'
            '  <base>dc=iinov,dc=com</base>'
            '  <user>'
            '    <dn>uid=_UID_,ou=people,dc=iinov,dc=com</dn>'
            '    <profiles>'
            '      <profile>profile_viewer</profile>'
            '      <profile>user_viewer</profile>'
            '    </profiles>'
            '  </user>'
            '</ldap>')
        self.assertIsNone(DBLdap.xml2db(dbsession, ldap_elt))
        dbldap = dbsession.query(DBLdap).first()
        self.assertIsInstance(dbldap, DBLdap)
        self.assertEqual(dbldap.host, 'localhost')
        self.assertEqual(dbldap.base, 'dc=iinov,dc=com')
        self.assertEqual(dbldap.user_dn, 'uid=_UID_,ou=people,dc=iinov,dc=com')
        user_profiles = [k.profile_id for k in dbldap.user_profiles]
        self.assertIn('profile_viewer', user_profiles)
        self.assertIn('user_viewer', user_profiles)

    # -------------------------------------------------------------------------
    def test_record_from_xml(self):
        """[u:modules.cioldap.models.dbldap.DBLdap.record_from_xml]"""
        from lxml import etree
        from ..models.dbldap import LDAP_CHECK_INTERVAL

        ldap_elt = etree.XML(
            '<ldap xmlns="http://ns.chrysal.io/cioldap">'
            '  <ssl>true</ssl>'
            '  <dn>cn=admin,dc=iinov,dc=com</dn>'
            '  <password>sesame</password>'
            '  <base>dc=iinov,dc=com</base>'
            '  <user>'
            '    <dn>uid=_UID_,ou=people,dc=iinov,dc=com</dn>'
            '    <last-name>sn</last-name>'
            '    <profiles>'
            '      <profile>profile_viewer</profile>'
            '    </profiles>'
            '  </user>'
            '</ldap>')
        record = DBLdap.record_from_xml(ldap_elt)
        self.assertIn('host', record)
        self.assertIn('port', record)
        self.assertEqual(record['port'], 389)
        self.assertIn('ssl', record)
        self.assertTrue(record['ssl'])
        self.assertIn('check_interval', record)
        self.assertEqual(record['check_interval'], LDAP_CHECK_INTERVAL)

    # -------------------------------------------------------------------------
    def test_record_format(self):
        """[u:modules.cioldap.models.dbldap.DBLdap.record_format]"""

        # ldap_id, root_password without root_dn, no base
        record = {'host': None, 'root_password': 'sesame'}
        error = DBLdap.record_format(record)
        self.assertNotIn('host', record)
        self.assertNotIn('root_password', record)
        self.assertIsNotNone(error)
        self.assertIn('No base', error)

        # root_password with root_dn, without user_dn
        record = {
            'root_dn': 'cn=admin,dc=iinov,dc=com', 'root_password': 'sesame',
            'base': 'dc=iinov,dc=com'}
        error = DBLdap.record_format(record)
        self.assertIn('root_password', record)
        self.assertTrue(record['root_password'].endswith('='))
        self.assertIsNotNone(error)
        self.assertIn('No DN', error)

        # user_dn without _UID_
        record['user_dn'] = 'uid=_FOO_,ou=people,dc=iinov,dc=com'
        error = DBLdap.record_format(record)
        self.assertIsNotNone(error)
        self.assertIn('_UID_', error)

        # Ok
        record['user_dn'] = 'uid=_UID_,ou=people,dc=iinov,dc=com'
        self.assertIsNone(DBLdap.record_format(record))

    # -------------------------------------------------------------------------
    def test_db2xml(self):
        """[u:modules.cioldap.models.dbldap.DBLdap.db2xml]"""
        dbldap = self._make_one()
        ldap_elt = dbldap.db2xml()
        self.assertEqual(ldap_elt.findtext('base'), 'dc=iinov,dc=com')
        self.assertEqual(
            ldap_elt.findtext('user/profiles/profile'), 'profile_viewer')

    # -------------------------------------------------------------------------
    def test_attachments2directory(self):
        """[u:modules.cioldap.models.dbldap.DBLdap.attachments2directory]"""
        from os.path import exists
        from ....tests import ATTACHMENTS_DIR, TEST_DIR

        DBLdap.attachments2directory(ATTACHMENTS_DIR, TEST_DIR)
        self.assertFalse(exists(TEST_DIR))

    # -------------------------------------------------------------------------
    def test_sheet4view(self):
        """[u:modules.cioldap.models.dbldap.DBLdap.sheet4view]"""
        from ....lib.form import Form

        profile_labels = {'profile_viewer': 'Profile viewer'}
        dbldap = self._make_one()
        html = dbldap.sheet4view(
            self.request, Form(self.request), profile_labels)
        self.assertIn('Profile viewer', html)

    # -------------------------------------------------------------------------
    def test_settings_schema(self):
        """[u:modules.cioldap.models.dbldap.DBLdap.settings_schema]"""
        from colander import SchemaNode
        from ..models.dbldap import LDAP_CHECK_INTERVAL

        profiles = {
            'profile_viewer': 'Profile viewer', 'user_viewer': 'User viewer'}

        # Creation mode
        schema, defaults = DBLdap.settings_schema(profiles)
        self.assertIsInstance(schema, SchemaNode)
        serialized = schema.serialize()
        self.assertIn('host', serialized)
        self.assertIn('user_dn', serialized)
        self.assertIn('pfl:profile_viewer', serialized)
        self.assertIsInstance(defaults, dict)
        self.assertIn('host', defaults)
        self.assertIn('port', defaults)
        self.assertIn('check_interval', defaults)
        self.assertEqual(defaults['check_interval'], LDAP_CHECK_INTERVAL)

        # Edition mode
        schema, defaults = DBLdap.settings_schema(profiles, self._make_one())
        self.assertIsInstance(defaults, dict)
        self.assertIn('pfl:profile_viewer', defaults)
        self.assertNotIn('pfl:user_viewer', defaults)

    # -------------------------------------------------------------------------
    def test_sheet4edit(self):
        """[u:modules.cioldap.models.dbldap.DBLdap.sheet4edit]"""
        from ....lib.form import Form

        profiles = {
            'profile_viewer': 'Profile viewer', 'user_viewer': 'User viewer'}
        dbldap = self._make_one()
        html = dbldap.sheet4edit(self.request, Form(self.request), profiles)
        self.assertIn('Attributed Profiles:', html)



# pylint: disable = import-outside-toplevel
"""Tests of ``security`` classes and functions."""

# pylint: disable = unused-import
from ..models.dbgroup import DBGroup  # noqa
# pylint: enable = unused-import
from . import ConfiguratorTestCase, DBUnitTestCase


# =============================================================================
class USecurityIncludeme(ConfiguratorTestCase):
    """Unit test class for :func:`subscribers.includeme`."""

    # -------------------------------------------------------------------------
    def test_include_security(self):
        """[u:security.includeme]"""
        self.configurator.include('..security')
        authorization_policy = self.configurator.introspector.get(
            'authorization policy', None)
        self.assertIsNotNone(authorization_policy)

        authentication_policy = self.configurator.introspector.get(
            'authentication policy', None)
        self.assertIsNotNone(authentication_policy)

        root_factory = self.configurator.introspector.get(
            'root factories', None).get('factory')
        self.assertGreater(len(root_factory.__acl__), 6)


# =============================================================================
class USecuritySessionAuthenticationPolicy(DBUnitTestCase):
    """Unit test class for :class:`subscribers.SessionAuthenticationPolicy`."""

    # -------------------------------------------------------------------------
    def test_authenticated_userid(self):
        """[u:security.SessionAuthenticationPolicy.authenticated_userid]"""
        from ..security import SessionAuthenticationPolicy

        authentication_policy = SessionAuthenticationPolicy(
            secret='-', hashalg='sha512', cookie_name='CIO_AUTH')

        user_id = authentication_policy.authenticated_userid(self.request)
        self.assertIsNone(user_id)

        self.add_user({
            'login': 'test1', 'last_name': 'Mr Test 1', 'password': 'test1pwd',
            'email': 'test1@chrysal.io'})
        self.configurator.testing_securitypolicy('test1', permissive=True)
        user_id = authentication_policy.authenticated_userid(self.request)
        self.assertEqual(user_id, 1)

        self.request.session['user'] = {'login': 'test1', 'user_id': 2}
        user_id = authentication_policy.authenticated_userid(self.request)
        self.assertEqual(user_id, 2)

        self.add_user({
            'login': 'test2', 'last_name': 'Mr Test 2', 'password': 'test2pwd',
            'email': 'test2@chrysal.io', 'password_mustchange': True})
        self.configurator.testing_securitypolicy('test2', permissive=True)
        self.assertIsNone(
            authentication_policy.authenticated_userid(self.request))

    # -------------------------------------------------------------------------
    def test_effective_principals(self):
        """[u:security.SessionAuthenticationPolicy.effective_principals]"""
        from ..security import SessionAuthenticationPolicy

        authentication_policy = SessionAuthenticationPolicy(
            secret='-', hashalg='sha512', cookie_name='CIO_AUTH')

        principals = authentication_policy.effective_principals(self.request)
        self.assertEqual(len(principals), 1)
        self.assertEqual(principals[0], 'system.Everyone')

        self.add_user({
            'login': 'test2', 'last_name': 'Mr Test 2', 'password': 'test2pwd',
            'email': 'test2@chrysal.io'})
        self.configurator.testing_securitypolicy('test2', permissive=True)
        principals = authentication_policy.effective_principals(self.request)
        self.assertEqual(len(principals), 2)
        self.assertEqual(principals[1], 'system.Authenticated')



"""Database main objects, functions and constants."""

from os import makedirs
from os.path import join, exists
from shutil import copy2
from json import loads

from ..lib.i18n import translate_field


# =============================================================================
class DBBaseClass(object):
    """Base class for prime SQLAlchemy table."""

    suffix = 'cio'
    attachments_dir = None
    _settings_tabs = None

    # -------------------------------------------------------------------------
    def label(self, request):
        """Return a translated label.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: str
        """
        # pylint: disable = no-member
        return translate_field(request, loads(self.i18n_label)) \
            if hasattr(self, 'i18n_label') else ''

    # -------------------------------------------------------------------------
    def description(self, request):
        """Return a translated description.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: str
        """
        # pylint: disable = no-member
        return translate_field(request, self.i18n_description) \
            if hasattr(self, 'i18n_description') else ''

    # -------------------------------------------------------------------------
    def attachments2directory(self, attachments, directory):
        """Copy from attachments directory the file corresponding to the user.

        :param str attachments:
            Absolute path to the attachments directory.
        :param str directory:
            The backup directory.
        """
        # pylint: disable = no-member
        if not self.attachments_dir or \
           not hasattr(self, 'attachments_key') or \
           not self.attachments_key or \
           not hasattr(self, 'picture') or not self.picture:
            return

        picture = join(
            attachments, self.attachments_dir, self.attachments_key,
            self.picture)
        if not exists(picture):
            return

        destination = join(
            directory, self.attachments_dir, self.attachments_key)
        if not exists(destination):
            makedirs(destination)
        copy2(picture, destination)

    # -------------------------------------------------------------------------
    @classmethod
    def settings_tabs(cls, request):
        """Return a tuple of tab labels to view or edit the object.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: tuple
        """
        # pylint: disable = unused-argument
        return cls._settings_tabs



# -*- coding: utf-8 -*-
"""Function to manage e-mails."""

from re import match
from smtplib import SMTP, SMTP_SSL
from smtplib import SMTPSenderRefused, SMTPRecipientsRefused, SMTPConnectError
from smtplib import SMTPHeloError, SMTPAuthenticationError, SMTPException
from socket import error as socket_error
from email import encoders
from email.charset import QP, Charset
from email.mime.nonmultipart import MIMENonMultipart
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.utils import formatdate

from chameleon import PageTemplateFile, PageTextTemplateFile

from pyramid.asset import abspath_from_asset_spec
from pyramid.i18n import TranslationString

from .i18n import _
from .log import log_error


# =============================================================================
class Mailing(object):
    """Class to send one or several mails.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        self._request = request
        self._domain = 'chrysalio'

    # -------------------------------------------------------------------------
    @classmethod
    def is_valid(cls, email):
        """Check the validity of an email address.

        :param str email:
            Address to check.
        :rtype: bool
        """
        return match(
            r'^[a-zA-Z0-9._%-]+@[a-zA-Z0-9._%-]+\.[a-zA-Z]{2,6}$', email)

    # -------------------------------------------------------------------------
    def send(self, email):
        """Send an email.

        :type  email: email.message.Message
        :param email:
            MIME type object to send.
        :rtype: :class:`str` or ``None``
        """
        settings = self._request.registry.settings
        try:
            if settings.get('smtp.ssl') == 'true':
                smtp = SMTP_SSL(
                    settings.get('smtp.host', 'localhost'),
                    int(settings.get('smtp.port', 465)))
            else:
                smtp = SMTP(
                    settings.get('smtp.host', 'localhost'),
                    int(settings.get('smtp.port', 25)))
        except SMTPConnectError as error:  # pragma: nocover
            log_error(self._request, error)
            return error
        except socket_error as error:
            log_error(self._request, error.strerror or error)
            return error.strerror or error

        if settings.get('smtp.user') and settings.get('smtp.password'):
            try:
                smtp.login(settings['smtp.user'], settings['smtp.password'])
            except (SMTPHeloError, SMTPAuthenticationError,
                    SMTPException) as error:
                smtp.quit()
                log_error(self._request, str(error))
                return str(error)

        try:
            smtp.sendmail(email['From'], email['To'], email.as_string())
        except (SMTPSenderRefused,
                SMTPRecipientsRefused) as error:  # pragma: nocover
            log_error(self._request, error)
            return error
        finally:
            smtp.quit()
        return None

    # -------------------------------------------------------------------------
    def mailing(self, email_template, recipients):
        """Send an e-mail to several users according to a HTML and a text
        template.

        :param dict email_template:
            A dictionary containing the e-mail parameters. The keys of this
            dictionary are: ``subject``, ``from``, ``text_template``,
            ``html_tempalte``, ``attachments``. ``subject`` can be overridden
            by the user dictionary.
        :param list recipients:
            A list of recipients to whom to send the e-mail. Each recipient is
            a dictionary containing at least the ``to`` key. It is passed to
            the Chameleon ``html`` and ``text`` templates.
        :rtype: tuple
        :return:
            List of errors.
        """
        # Prepare templates
        templates = {}
        if 'html_template' in email_template:
            path = email_template['html_template']
            templates['html'] = (
                PageTemplateFile(abspath_from_asset_spec(path)),
                ':' in path and path.partition(':')[0] or 'chrysalio')
        if 'text_template' in email_template:
            path = email_template['text_template']
            templates['text'] = (
                PageTextTemplateFile(abspath_from_asset_spec(path)),
                ':' in path and path.partition(':')[0] or 'chrysalio')
        if not templates:
            log_error(self._request, 'No available template.')
            return (_('No available template.'),)
        translate = self._request.localizer.translate
        email_template['subject'] = translate(email_template['subject'])
        email_template['locale_name'] = self._request.locale_name
        email_template['_'] = self._template_translate

        # Browse recipients
        errors = set()
        for recipient in recipients:
            context = email_template.copy()
            context.update(recipient)
            if 'attachments' in email_template or 'attachments' in recipient:
                email = MIMEMultipart()
                email.preamble = \
                    'This is a multi-part message in MIME format.\n'
                email.attach(self._mime_text(templates, context))
                for item in tuple(recipient.get('attachments', '')) + tuple(
                        email_template.get('attachments', '')):
                    part = MIMEBase('application', 'octet-stream')
                    with open(item[1], 'rb') as hdl:
                        part.set_payload(hdl.read())
                        encoders.encode_base64(part)
                    part.add_header(
                        'Content-Disposition',
                        'attachment; filename= {0}'.format(item[0]))
                    email.attach(part)
            else:
                email = self._mime_text(templates, context)

            email['Subject'] = context['subject']
            email['From'] = email_template['from']
            email['To'] = recipient['to']
            email['Date'] = formatdate()
            errors.add(self.send(email))

        return errors and errors != set((None,)) and tuple(errors)

    # -------------------------------------------------------------------------
    def _mime_text(self, templates, context):
        """Return a simple or a multi-part MIME object according to
        ``templates``.

        :param dict templates:
            A dictionary with keys ``text`` and/or ``html``.
        :param dict context:
            Parameters for templates.
        :rtype: :class:`email.mime.multipart.MIMEMultipart` or
            :class:`email.mime.text.MIMETest`
        """
        if len(templates) > 1:
            mime = MIMEMultipart('alternative')
            mime.preamble = 'This is a multi-part message in MIME format.\n'
            self._domain = templates['text'][1]
            mime.attach(self._mime_quoted_printable(
                templates['text'][0](**context)))
            self._domain = templates['html'][1]
            mime.attach(self._mime_quoted_printable(
                templates['html'][0](**context), 'html'))

        elif 'html' in templates:
            self._domain = templates['html'][1]
            mime = self._mime_quoted_printable(
                templates['html'][0](**context), 'html')

        else:
            self._domain = templates['text'][1]
            mime = self._mime_quoted_printable(
                templates['text'][0](**context))

        # Save HTML part in a file
        # if 'html' in templates:
        #     from os.path import dirname, join
        #     self._domain = templates['html'][1]
        #     with open(join(dirname(__file__),
        #                    '..', '..', 'email~.html'), 'w') as hdl:
        #         hdl.write(templates['html'][0](**context))

        return mime

    # -------------------------------------------------------------------------
    @classmethod
    def _mime_quoted_printable(cls, payload, subtype='plain'):
        """Encode MIMEText as quoted printables.

        :param str payload:
            Text to encode
        :param str subtype: (default='plain')
            Subtype for the text.
        :rtype: email.mime.nonmultipart.MIMENonMultipart
        """
        charset = Charset('utf-8')
        charset.header_encoding = QP
        charset.body_encoding = QP
        mime = MIMENonMultipart('text', subtype, charset='utf-8')
        mime.set_payload(payload, charset)
        return mime

    # -------------------------------------------------------------------------
    def _template_translate(self, text, mapping=None, domain=None):
        """Translation from a string for templating system."""
        return self._request.localizer.translate(TranslationString(
            text, mapping=mapping, domain=domain or self._domain))



# -*- coding: utf-8 -*-
"""Some various utilities."""

# pylint: disable = unused-import
try:  # pragma: nocover
    from os import scandir, walk
except ImportError:  # pragma: nocover
    from scandir import scandir, walk  # noqa
# pylint: enable = unused-import
from os import makedirs
from os.path import basename, exists, join, isdir, getmtime, dirname
from os.path import commonprefix
from sys import version_info
from logging import getLogger
from shutil import copy2
from unicodedata import normalize, combining
from base64 import b64encode, b64decode
from hashlib import sha256, md5
from random import SystemRandom, randrange
from string import ascii_uppercase, digits
from re import UNICODE, sub as re_sub
from binascii import Error as BAError
from datetime import datetime
from filecmp import cmpfiles
from mimetypes import guess_type
from subprocess import PIPE, STDOUT, Popen, TimeoutExpired
from subprocess import SubprocessError
from bcrypt import hashpw, gensalt

from future.moves.urllib.parse import urlparse, urlunparse
from Crypto import Random
from Crypto.Cipher import AES
from docutils.core import publish_parts

from pyramid.asset import abspath_from_asset_spec

from .i18n import _, translate


LOG = getLogger(__name__)
EXCLUDED_FILES = (
    '.git', '.hg', '.svn', 'Thumbs.db', '.DS_Store', '_MACOSX', '__MACOSX',
    '__pycache__')


# =============================================================================
def tostr(text):
    """Make a conversion according to Python version.

    :type  text: :class:`str` or :class:`bytes`
    :param text:
        Text to convert.
    :rtype: str
    """
    return \
        (version_info > (3, 0) and isinstance(text, bytes) and
         text.decode('utf8')) or \
        (version_info > (3, 0) and text) or \
        text.encode('utf8')


# =============================================================================
def tounicode(text):
    """Make a conversion according to Python version.

    :type  text: :class:`str` or :class:`bytes`
    :param text:
        Text to convert.
    :rtype: :class:`str` (Python 3) or ``unicode`` (Python 2)
    """
    return \
        (version_info > (3, 0) and isinstance(text, bytes) and
         text.decode('utf8')) or \
        (version_info > (3, 0) and text) or \
        (isinstance(text, (str, bytes)) and text.decode('utf8')) or text


# =============================================================================
def load_guessing_encoding(filename):
    """Tries to open a file by guessing its encoding.

    :param str filename:
        Absolute path to the file.
    :rtype: :class:`str` (Python 3) or ``unicode`` (Python 2) or ``None``
    """
    guessed = False
    with open(filename, 'rb') as hdl:
        content = hdl.read()

    # UTF-8
    try:
        content = content.decode('utf_8')
        guessed = True
    except UnicodeDecodeError:
        pass

    # Latin 1
    if not guessed:
        try:
            content = content.decode('latin_1')
            guessed = True
        except UnicodeDecodeError:  # pragma: nocover
            pass

    return content if guessed else None


# =============================================================================
def copy_content(src_dir, dst_dir, exclude=None, force=False):
    """Copy the content of a ``src_dir`` directory into a ``dst_dir``
    directory.

    :param str src_dir:
        Source directory path.
    :param str dst_dir:
        Destination directory path.
    :param list exclude: (optional)
        List of files to exclude.
    :param bool force: (optional)
        Force copy even if the target file has the same date.
    """
    if not exists(dst_dir):
        makedirs(dst_dir)
    if exclude is None:
        exclude = EXCLUDED_FILES
    for entry in scandir(src_dir):
        if entry.name not in exclude:
            destination = join(dst_dir, tounicode(entry.name))
            if entry.is_dir():
                copy_content(entry.path, destination, exclude, force)
            elif force or not exists(destination) \
                    or getmtime(destination) != getmtime(entry.path):
                copy2(entry.path, destination)


# =============================================================================
def copy_content_re(src_dir, dst_dir, exclude=None):
    """Copy the content of a ``src_dir`` directory into a ``dst_dir``
    directory.

    :param str src_dir:
        Source directory path.
    :param str dst_dir:
        Destination directory path.
    :param exclude: (optional)
        Regular expression to exclude files during the copy.
    """
    if not exists(dst_dir):
        makedirs(dst_dir)
    for entry in scandir(src_dir):
        if exclude is None or exclude.search(entry.name) is None:
            destination = join(dst_dir, entry.name)
            if entry.is_dir():
                copy_content_re(entry.path, destination, exclude)
            else:
                copy2(entry.path, destination)


# =============================================================================
def make_id(name, mode=None, truncate=None):
    """Make an ID with name.

    :param str name:
        Name to use.
    :param int truncate: (optional)
        If not ``None``, maximum length of the returned string.
    :param str mode: (optional)
        Strategy to make ID : 'standard', 'token', 'xmlid', 'class' or
        'no_accent'.
    :rtype: str

    Examples of transformation of `12Test___Té*t.;?!`:

    * mode = ``None``: `12test___té*t.;?!`
    * mode = 'standard': `12Test_Té_t._`
    * mode = 'token': `12test_te_t._`
    * mode = 'xmlid': `_12test_te_t._`
    * mode = 'class': `12Test_Te_t_`
    * mode = 'no_accent': `12Test___Te*t.;?!`
    """
    result = name.strip()
    if mode not in ('standard', 'class', 'no_accent'):
        result = result.lower()
    if mode in ('standard', 'token', 'xmlid', 'class'):
        result = re_sub(
            '_+', '_', re_sub('[  *?!;:,"\'/«»()\\[\\]–&]', '_', result))
    if mode in ('token', 'xmlid', 'class', 'no_accent'):
        result = normalize('NFKD', result.encode('utf8').decode('utf8'))
        result = ''.join([k for k in result if not combining(k)])
    if mode == 'xmlid' and result and result[0].isdigit():
        result = '_{0}'.format(result)
    if mode == 'class':
        result = re_sub('_+', '_', re_sub(r'\.', '_', result))
    return result[0:truncate] if truncate else result


# =============================================================================
def make_digest(name):
    """Create a digest key with name.

    :param str name:
        Name to use.
    """
    return md5(name.encode('utf8')).hexdigest()


# =============================================================================
def normalize_spaces(text, truncate=None):
    """Normalize spaces and, possibly, truncate result.

    :param str text:
        Text to normalize.
    :param int truncate: (optional)
        If not ``None``, maximum lenght of the returned string.
    :rtype: :class:`str` or ``None``
    """
    if text is None:
        return None
    text = ' '.join(text.replace(' ', ':~:').strip().split())\
        .replace(':~:', ' ')
    return text[0:truncate] if truncate else text


# =============================================================================
def camel_case(text):
    """Convert ``text`` in Camel Case.

    :param str text:
        Text to transform.

    Examples of transformation:

    xml2html -> Xml2Html
    laTeX -> LaTeX
    my_way -> MyWay
    my way -> MyWay
    my-way -> MyWay
    """
    return re_sub(
        r'(^\w|[-_ 0-9]+\w)',
        lambda m: m.group(0).replace('_', '').replace(' ', '').upper(),
        text, flags=UNICODE)


# =============================================================================
def shorten(text, width, placeholder='…'):
    """Collapse and truncate the given text to fit in the given width.

    :param str text:
        Text to shorten.
    :param int with:
        The result fits in the width.
    :param str placeholder: (default=…)
        Place holder to use if the size is too big.
    :rtype: str
    """
    text = normalize_spaces(text)
    return text if len(text) <= width else \
        '{0}{1}'.format(text[:width - len(placeholder)], placeholder)


# =============================================================================
def encrypt(value, key, init_vector=None):
    """Encryption function.

    :param bytes value:
        Value to encrypt.
    :param str key:
        Encryption key.
    :param bytes init_vector: (optional)
        Initialization vector for AES algorithm.
    :rtype: bytes
    :retrun:
        Encrypted value or ``None``.
    """
    if not value:
        return value
    bsize = AES.block_size
    if version_info > (3, 0) and not isinstance(value, bytes):
        value = bytes(value, 'utf8')  # pragma: nocover
    value = value + \
        (bsize - len(value) % bsize) * \
        chr(bsize - len(value) % bsize).encode('latin-1')
    if not isinstance(key, bytes):
        key = key.encode('utf8')
    key = sha256(key).digest()
    init_vector = init_vector or Random.new().read(AES.block_size)
    cipher = AES.new(key, AES.MODE_CBC, init_vector)
    return tostr(b64encode(init_vector + cipher.encrypt(value)))


# =============================================================================
def decrypt(value, key):
    """Encryption function with padding.

    :param bytes value:
        String to encrypt.
    :param str key:
        Encryption key.
    :param str init_vector: (optional)
        Initialization vector for AES algorithm.
    :rtype: str
    :return:
        Encrypted value or ``None``.
    """
    if not value:
        return value
    try:
        value = b64decode(value)
    except (TypeError, BAError):
        return None
    if not isinstance(key, bytes):
        key = key.encode('utf8')
    key = sha256(key).digest()
    init_vector = value[:AES.block_size]
    if len(init_vector) != AES.block_size:
        return None
    cipher = AES.new(key, AES.MODE_CBC, init_vector)
    try:
        value = cipher.decrypt(value[AES.block_size:])
    except ValueError:  # pragma: nocover
        return None
    try:
        return value[:-ord(value[len(value) - 1:])].decode('utf8')
    except UnicodeDecodeError:  # pragma: nocover
        return None


# =============================================================================
def token(length=None):
    """Generate a token of length ``length`` or with a length between 8 an 16
    characters.

    :param int length: (optional)
        Length of the token.
    :rtype: str
    :return:
        Token.
    """
    if length is None:
        length = randrange(8, 16)
    return ''.join(
        SystemRandom().choice(ascii_uppercase + digits) for _ in range(length))


# =============================================================================
def execute(command, cwd=None, no_exit_code=False, timeout=None, **kwargs):
    """Run the command described by command. Wait for command to complete.
    If the return code is not zero, return output and an error message.

    :param list command:
        Splitted command to execute.
    :param str cwd: (optional)
        If it is not ``None``, the current directory will be changed to ``cwd``
        before it is executed.
    :param bool no_exit_code: (default=False)
        If the command is known to exit with code 0 even if there is an error,
        assign this argument to ``True``.
    :param float timeout: (default=None)
        If set and the process hasn't finished in that time (in seconds),
        exits with error. The process is not killed.
    :param dict kwargs: (optional)
        Dictionary of keyword arguments.
    :rtype: tuple
    :return:
        An error message such as ``(output, error)`` where ``output`` is a
        string and ``error`` a :class:`pyramid.i18n.TranslationString`.
    """
    # pylint: disable = consider-using-with
    try:
        process = Popen(command, cwd=cwd, stderr=STDOUT, stdout=PIPE, **kwargs)
    except (OSError, ValueError, SubprocessError) as error:
        return '', _('"${c}" failed: ${e}', {'c': command, 'e': error})
    if command[0] == 'nice':
        command = command[1:]
    command = basename(command[0])
    try:
        output = process.communicate(timeout=timeout)[0]
        if process.poll() or (no_exit_code and output):
            try:
                return output[0:102400].decode('utf8').strip(), \
                    _('"${c}" failed', {'c': command})
            except UnicodeDecodeError:  # pragma: nocover
                return output[0:102400].decode('latin1').strip(), \
                    _('"${c}" failed', {'c': command})
    except (OSError, TimeoutExpired) as error:  # pragma: nocover
        return '', _('"${c}" failed: ${e}', {'c': command, 'e': error})
    output = output[0:102400]
    try:
        output = output.decode('utf8')
    except UnicodeDecodeError:  # pragma: nocover
        pass
    return output.strip(), ''


# =============================================================================
def full_url(url, user=None, password=None):
    """Return an URL with ``user`` and possibly ``password``.

    :param str url:
        URL to update.
    :param str user:
        User for VCS access.
    :param str password:
        Password for VCS access.
    :rtype: str
    :return:
        Full URL.
    """
    if not user:
        return url
    scheme, netloc, path, params, query, fragment = urlparse(str(url))
    netloc = '{0}{1}@{2}'.format(
        user, password and ':{0}'.format(password) or '',
        netloc.rpartition('@')[2])
    return urlunparse((scheme, netloc, path, params, query, fragment))


# =============================================================================
def mimetype_get(full_path):
    """Return the mime type of ``full_path``.

    :param str full_path:
        Absolute path to the file.
    :rtype: tuple
    :return:
        A tuple such as ``(mimetype, subtype)``. For instance:
        ``('text/plain', 'plain')``.
    """
    if isdir(full_path):
        return 'directory', 'directory'
    mimetype = guess_type(tostr(full_path), False)[0]
    if mimetype is None:
        return 'unknown', 'unknown'
    subtype = mimetype.partition('/')[2]
    return mimetype, subtype or mimetype


# =============================================================================
def deltatime_label(seconds=0, minutes=0, hours=0, days=0, lang=None):
    """Return a translated label for a delta time.

    :param int seconds:
        Number of seconds.
    :param int minutes:
        Number of minutes.
    :param int hours:
        Number of hours.
    :param int days:
        Number of days.
    :param str lang: (optional)
        Language to use.
    :rtype: pyramid.i18n.TranslationString
    :return:
        Return a human reading translation string.
    """
    delta = seconds + 60 * minutes + 3600 * hours + 86400 * days
    if delta <= 0:
        return translate(_('0 second'), lang)

    days = int(delta // 86400)
    delta = delta % 86400
    hours = int(delta // 3600)
    delta = delta % 3600
    minutes = int(delta // 60)
    seconds = delta % 60
    label = ''

    if days:
        label = translate(
            _('1 day') if days == 1 else
            _('${d} days', {'d': days}), lang)
    if hours:
        label = '{0} '.format(label) if label else ''
        label += translate(
            _('1 hour') if hours == 1 else
            _('${h} hours', {'h': hours}), lang)
    if minutes:
        label = '{0} '.format(label) if label else ''
        label += translate(
            _('1 minute') if minutes == 1 else
            _('${m} minutes', {'m': minutes}), lang)
    if seconds:
        label = '{0} '.format(label) if label else ''
        label += translate(
            _('1 second') if seconds == 1 else
            _('${s} seconds', {'s': seconds}), lang)

    return label


# =============================================================================
def age(mtime):
    """Return an age in minutes, hours, days or a date.

    :param datetime mtime:
        Modification time.
    :rtype: pyramid.i18n.TranslationString
    :return:
        Return an age or a date if ``mtime`` is older than a year.
    """
    # pylint: disable = too-many-return-statements
    if not mtime:
        return ''
    delta = datetime.now() - mtime

    if delta.days < 0:
        return _('0 second')
    if delta.days == 0 and delta.seconds < 60:
        return _('1 second') if delta.seconds <= 1 \
            else _('${s} seconds', {'s': delta.seconds})
    if delta.days == 0 and delta.seconds < 3600:
        minutes = int(round(delta.seconds / 60))
        return _('1 minute') if minutes == 1 \
            else _('${m} minutes', {'m': minutes})
    if delta.days == 0:
        hours = int(round(delta.seconds / 3600))
        return _('1 hour') if hours == 1 \
            else _('${h} hours', {'h': hours})
    if delta.days < 7:
        return _('1 day') if delta.days == 1 \
            else _('${d} days', {'d': delta.days})
    if delta.days < 30:
        weeks = int(round(delta.days / 7))
        return _('1 week') if weeks == 1 \
            else _('${w} weeks', {'w': weeks})
    if delta.days < 365:
        months = int(round(delta.days / 30))
        return _('1 month') if months == 1 else \
            _('${m} months', {'m': months})

    return str(mtime.replace(microsecond=0))[0:-9]


# =============================================================================
def size_label(size, is_dir=False):
    """Return a size in o, Kio, Mio or Gio.

    :param int size:
        Size in figures.
    :param bool is_dir: (optional)
        ``True`` if it is about a directory.
    :rtype: :class:`str` or :class:`pyramid.i18n.TranslationString`
    """
    # For a directory
    if is_dir:
        return _('${n} items', {'n': size}) if size > 1 else \
            _('${n} item', {'n': size})

    # For a file
    if size >= 1073741824:
        return '{0:.1f} Gio'.format(round(size / 1073741824.0, 1))
    if size >= 1048576:
        return '{0:.1f} Mio'.format(round(size / 1048576.0, 1))
    if size >= 1024:
        return '{0:.1f} Kio'.format(round(size / 1024.0, 1))
    return '{0} o'.format(size)


# =============================================================================
def convert_value(type_, value):
    """Convert a string value according to type ``type_``.

    :param str type_:
        Type of the value.
    :rtype: :class:`str`, :class:`bool`, :class:`int`, :class:`float` or
        :class:`datetime.date`
    """
    if type_ == 'boolean':
        value = value not in (None, '', 'false', 'False', '0')
    elif type_ == 'integer':
        value = int(value) if value and value.isdigit() else 0
    elif type_ == 'decimal':
        try:
            value = float(value)
        except ValueError:
            value = 0
    elif type_ == 'datetime':
        try:
            value = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S').replace(
                microsecond=0, tzinfo=None)
        except ValueError:
            try:
                value = datetime.strptime(value, '%Y-%m-%d')
            except ValueError:
                value = None
    elif type_ == 'date':
        try:
            value = datetime.strptime(value, '%Y-%m-%d').date()
        except ValueError:
            value = None
    return value


# =============================================================================
def check_chrysalio_js(js_dir):
    """Check if the ``chrysalio.js`` file exists in ``js_dir`` directory and if
    it is the last version and possibly update them.

    :param str js_dir:
        Directory for Javascript files.
    """
    js_dir = abspath_from_asset_spec(js_dir)
    if not exists(js_dir):
        return

    cio_js_dir = join(dirname(__file__), '..', 'Static', 'Js')
    js_files = (
        'jquery.js', 'js.cookie.js', 'jquery-ui.js',
        'jquery-ui-datepicker-fr.js', 'chrysalio.js',
        'chrysalio-datepicker.js', 'chrysalio-filter.js')
    for name in cmpfiles(cio_js_dir, js_dir, js_files)[1]:
        try:
            copy2(join(cio_js_dir, name), js_dir)
        except IOError:  # pragma: nocover
            LOG.warning('"%s" is not up to date.', name)


# =============================================================================
def check_chrysalio_css(css_dir):
    """Check if the Chrysalio CSS file exists in ``css_dir`` directory
    and if it is the last version and possibly update them.

    :param str css_dir:
        Directory for CSS files.
    """
    css_dir = abspath_from_asset_spec(css_dir)
    if not exists(css_dir):
        return
    cio_css_dir = join(dirname(__file__), '..', 'Static', 'Css')
    css_files = ('jquery-ui.css',)
    for name in cmpfiles(cio_css_dir, css_dir, css_files)[1]:
        try:
            copy2(join(cio_css_dir, name), css_dir)
        except IOError:  # pragma: nocover
            LOG.warning('"%s" is not up to date.', name)


# =============================================================================
def common_directory(files):
    """Return the common directory of the list of files.

    :param list files:
        List of files to analyse.
    :rtype: str
    """
    if not files:
        return None
    directory = commonprefix(files)
    if not isdir(directory):
        directory = dirname(directory)
    return directory if isdir(directory) else None


# =============================================================================
def hash_admin_password(project_dir, files, password):
    """Add hashed password for administrator in each file of the ``files`` list
    contained in ``project_dir`` directory.

    :param str project_dir:
        Project directory containing INI files.
    :param tuple files:
        List of files to process.
    :param str password:
        Clear password.
    """
    hashed = hashpw(password.encode('utf8'), gensalt()).decode('utf8')
    for filename in files:
        filename = join(project_dir, filename)
        with open(filename, 'r') as hdl:
            content = hdl.read()
        content = content.replace(
            'admin.password = {0}'.format(password),
            '#admin.password = {0}\nadmin.password = {1}'.format(
                password, hashed))
        with open(filename, 'w') as hdl:
            hdl.write(content)


# =============================================================================
def rst2html(rst):
    """Transform a reStructuredText into HTML.

    :param str rst:
        reStructuredText.
    :rtype: str
    :return:
        XHTML.
    """
    html = publish_parts(rst, writer_name='html')['body'] if rst else None
    return re_sub('(^<p>|</p>\n)', '', html) \
        if rst and 'ERROR/' not in html else rst



"""CIOSkeleton Relax NG."""

from os.path import dirname, join


RELAXNG_CIOSKELETON = {
    'namespace': 'http://ns.chrysal.io/cioskeleton',
    'root': 'cioskeleton', 'version': '1.0',
    'file': join(dirname(__file__), 'RelaxNG', 'cioskeleton.rng')}



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.mailing.Mailing`` class."""

from os.path import basename
from email.mime.text import MIMEText
from unittest import TestCase

from pyramid.testing import DummyRequest


# =============================================================================
class ULibMailingMailing(TestCase):
    """Unit test class for :class:`lib.mailing.Mailing`."""

    # -------------------------------------------------------------------------
    def test_is_valid(self):
        """[u:lib.mailing.Mailing.is_valid]"""
        from ..lib.mailing import Mailing

        self.assertTrue(Mailing.is_valid('foo@chrysal.io'))
        self.assertFalse(Mailing.is_valid('@chrysal.io'))
        self.assertFalse(Mailing.is_valid('foochrysal.io'))
        self.assertFalse(Mailing.is_valid('foo @ chrysal.io'))

    # -------------------------------------------------------------------------
    def test_send(self):
        """[u:lib.mailing.Mailing.send]"""
        from ..lib.utils import decrypt
        from ..lib.mailing import Mailing

        request = DummyRequest()
        request.registry.settings = {'smtp.host':  'chrysalio'}
        email = MIMEText('Hello World!')
        email['From'] = 'admin@chrysal.io'
        email['To'] = 'test@chrysal.io'
        email['Subject'] = '[Chrysalio] Hello'

        # Socket error
        error = Mailing(request).send(email)
        self.assertTrue(
            ('Connection refused' in error) or
            ('Name or service not known' in error) or
            ('No address associated with hostname' in error) or
            ('Temporary failure' in error) or
            ('Connection timed out' in error))

        # Socket error with SSL
        request.registry.settings['smtp.host'] = 'chrysalio'
        request.registry.settings['smtp.ssl'] = 'true'
        error = Mailing(request).send(email)
        self.assertTrue(
            ('Connection refused' in error) or
            ('Name or service not known' in error) or
            ('No address associated with hostname' in error) or
            ('Temporary failure' in error) or
            ('Connection timed out' in error))

        # Connection
        request.registry.settings['smtp.host'] = 'mail.gandi.net'
        request.registry.settings['smtp.ssl'] = 'true'
        request.registry.settings['smtp.user'] = 'test@chrysal.io'
        request.registry.settings['smtp.password'] = decrypt(
            'CQLanGWkUonUrXKYheoQBo0vEtoFMlGuklBS5wIqj/U=', 'warehouse')
        error = Mailing(request).send(email)
        if error:
            return

        # Bad authentication
        request.registry.settings['smtp.password'] = 'foo'
        error = Mailing(request).send(email)
        self.assertIn('authentication failed', error)

    # -------------------------------------------------------------------------
    def test_mailing(self):
        """[u:lib.mailing.Mailing.mailing]"""
        from . import EMAIL_HTML, EMAIL_TEXT, TEST1_SVG
        from ..lib.mailing import Mailing

        request = DummyRequest()
        request.registry.settings = {'smtp.host': 'chrysalio'}
        email_template = {'subject': 'Test', 'from': 'admin@chrysal.io'}
        recipient = {
            'to': 'test1@chrysal.io',
            'first_name': 'Édith',
            'last_name': 'AVULEUR'}

        errors = Mailing(request).mailing(email_template, [recipient])
        self.assertIsInstance(errors, tuple)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], 'No available template.')

        email_template['html_template'] = EMAIL_HTML
        errors = Mailing(request).mailing(email_template, [recipient])
        self.assertIsInstance(errors, tuple)
        self.assertEqual(len(errors), 1)
        self.assertTrue(
            ('Connection refused' in errors[0]) or
            ('Name or service not known' in errors[0]) or
            ('No address associated with hostname' in errors[0]) or
            ('Temporary failure' in errors[0]) or
            ('Connection timed out' in errors[0]))

        del email_template['html_template']
        email_template['text_template'] = EMAIL_TEXT
        errors = Mailing(request).mailing(email_template, [recipient])
        self.assertTrue(
            ('Connection refused' in errors[0]) or
            ('Name or service not known' in errors[0]) or
            ('No address associated with hostname' in errors[0]) or
            ('Temporary failure' in errors[0]) or
            ('Connection timed out' in errors[0]))

        email_template['html_template'] = EMAIL_HTML
        recipient['attachments'] = ((basename(TEST1_SVG), TEST1_SVG),)
        errors = Mailing(request).mailing(email_template, [recipient])
        self.assertTrue(
            ('Connection refused' in errors[0]) or
            ('Name or service not known' in errors[0]) or
            ('No address associated with hostname' in errors[0]) or
            ('Temporary failure' in errors[0]) or
            ('Connection timed out' in errors[0]))



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``views.login`` functions."""

from . import DBUnitTestCase, FunctionalTestCase


# =============================================================================
class UViewsLoginLogin(DBUnitTestCase):
    """Unit test class for testing :func:`views.login.login`."""

    users = {
        'ok': {
            'login': 'user1', 'first_name': 'Marc',
            'last_name': 'HINDÈLAIBILE', 'email': 'user1@chrysal.io',
            'password': 'user1pwd'},
        'inactive': {
            'login': 'test2', 'first_name': 'Sophie', 'last_name': 'FONFEC',
            'status': 'inactive', 'email': 'test2@chrysal.io',
            'password': 'test2pwd'},
        'locked': {
            'login': 'test3', 'first_name': 'Guy', 'last_name': 'LIGUILI',
            'status': 'locked', 'email': 'test3@chrysal.io',
            'password': 'test3pwd'},
        'expired': {
            'login': 'test4', 'first_name': 'Gédéon',
            'last_name': 'TEUSEMANIE', 'email': 'test4@chrysal.io',
            'password': 'test4pwd', 'expiration': '2018-01-01'},
        'password_change': {
            'login': 'test5', 'first_name': 'Sébastienne',
            'last_name': 'TOUSSEUL', 'email': 'test5@chrysal.io',
            'password': 'test4pwd', 'password_mustchange': True}}

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        super(UViewsLoginLogin, self).setUp()
        self.configurator.add_route('home', '/')
        self.configurator.add_route('login', '/login')
        self.configurator.add_route(
            'user_password_reset', '/user/password/reset/{user_id}/{token}')

    # -------------------------------------------------------------------------
    def fill_form(self, login, password):
        """Fill the login form.

        :param str login:
            Login value.
        :param str password:
            Clear password.
        """
        self.request.params = {'login': login, 'password': password}
        self.request.POST = self.request.params

    # -------------------------------------------------------------------------
    def test_without_login(self):
        """[u:views.login.login] without login"""
        from ..views.login import login
        from ..lib.form import Form

        url_from = 'http://localhost/foo'
        self.request.url = url_from
        response = login(self.request)
        self.assertIn('next', response)
        self.assertEqual(response['next'], url_from)
        self.assertIn('form', response)
        self.assertIsInstance(response['form'], Form)
        self.assertIn('next', response['form'].values)
        self.assertEqual(response['form'].values['next'], url_from)

    # -------------------------------------------------------------------------
    def test_empty_password(self):
        """[u:views.login.login] empty password"""
        from ..views.login import login

        self.fill_form(self.users['ok']['login'], None)
        self.add_user(self.users['ok'])

        response = login(self.request)
        alert = self.request.session.pop_flash('alert')
        self.assertFalse(alert)
        self.assertIn('form', response)

    # -------------------------------------------------------------------------
    def test_invalid_password(self):
        """[u:views.login.login] invalid password"""
        from ..views.login import login

        self.fill_form(self.users['ok']['login'], 'invalid_password')
        self.add_user(self.users['ok'])

        login(self.request)
        alert = self.request.session.pop_flash('alert')
        self.assertTrue(alert)
        self.assertIn('is incorrect', alert[0])

    # -------------------------------------------------------------------------
    def test_not_active(self):
        """[u:views.login.login] not active"""
        from ..views.login import login

        self.fill_form(
            self.users['inactive']['login'],
            self.users['inactive']['password'])
        self.add_user(self.users['inactive'])

        login(self.request)
        alert = self.request.session.pop_flash('alert')
        self.assertTrue(alert)
        self.assertIn('is not active', alert[0])

    # -------------------------------------------------------------------------
    def test_locked(self):
        """[u:views.login.login] locked"""
        from ..views.login import login

        self.fill_form(
            self.users['locked']['login'],
            self.users['locked']['password'])
        self.add_user(self.users['locked'])

        login(self.request)
        alert = self.request.session.pop_flash('alert')
        self.assertTrue(alert)
        self.assertIn('is locked', alert[0])

    # -------------------------------------------------------------------------
    def test_expired(self):
        """[u:views.login.login] expired"""
        from ..views.login import login

        self.fill_form(
            self.users['expired']['login'],
            self.users['expired']['password'])
        self.add_user(self.users['expired'])

        login(self.request)
        alert = self.request.session.pop_flash('alert')
        self.assertTrue(alert)
        self.assertIn('has expired', alert[0])

    # -------------------------------------------------------------------------
    def test_password_change(self):
        """[u:views.login.login] password change"""
        from pyramid.httpexceptions import HTTPFound
        from ..views.login import login

        self.fill_form(
            self.users['password_change']['login'],
            self.users['password_change']['password'])
        self.add_user(self.users['password_change'])

        response = login(self.request)
        self.assertIsInstance(response, HTTPFound)
        self.assertIn('password/reset', response.location)

    # -------------------------------------------------------------------------
    def test_authorized(self):
        """[u:views.login.login] authorized"""
        from pyramid.httpexceptions import HTTPFound
        from ..views.login import login

        self.fill_form(self.users['ok']['login'], self.users['ok']['password'])
        self.add_user(self.users['ok'])

        response = login(self.request)
        self.assertIsInstance(response, HTTPFound)


# =============================================================================
class FViewsLoginLogin(FunctionalTestCase):
    """Functional test class for testing :func:`views.login.login`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[f:views.login.login]"""
        response = self.testapp.get('/login', status=200)
        self.assertIn(b'Log in', response.body)


# =============================================================================
class UViewsLoginLogout(DBUnitTestCase):
    """Unit test class for testing :func:`views.login.logout`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:views.login.logout]"""
        from pyramid.httpexceptions import HTTPFound
        from ..views.login import logout

        self.configurator.add_route('login', '/login')
        self.request.session['user'] = {'user_id': 1, 'login': 'test1'}

        response = logout(self.request)
        self.assertIsInstance(response, HTTPFound)



# pylint: disable = import-outside-toplevel
"""Tests of ``routes`` classes and functions."""

from . import ConfiguratorTestCase


# =============================================================================
class URoutesIncludeme(ConfiguratorTestCase):
    """Unit test class for :func:`routes.includeme`."""

    # -------------------------------------------------------------------------
    def test_include_routes(self):
        """[u:routes.includeme]"""
        self.configurator.include('..routes')
        self.assertIsNotNone(
            self.configurator.registry.introspector.get('routes', 'home'))



"""Function to import and export database from and into XML files."""

from lxml import etree

from ....models.populate import element2db
from ..relaxng import RELAXNG_CIOSKELETON
from .dbbone import DBBone


# =============================================================================
def xml2db(dbsession, root_elt, only=None, error_if_exists=True, modules=None):
    """Load an XML configuration file for an included module.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  root_elt: lxml.etree.Element
    :param root_elt:
        XML element with the namespace of the module.
    :param str only: (optional)
        If not ``None``, only the items of type ``only`` are loaded.
    :param bool error_if_exists: (default=True)
        It returns an error if an item already exists.
    :type  modules: collections.OrderedDict
    :param modules: (optional)
        Dictionary of modules to use to complete the loading.
    :rtype: list
    :return:
        A list of error messages.
    """
    # pylint: disable = unused-argument
    return element2db(
        dbsession, root_elt, only, error_if_exists, {
            'tag': 'bone', 'class': DBBone, 'relaxng': RELAXNG_CIOSKELETON})


# =============================================================================
def db2xml(dbsession, root_elt):
    """Return a list of XML elements.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  root_elt: lxml.etree._Element
    :param root_elt:
        XML element with the namespace of the module.
    """
    dbbones = dbsession.query(DBBone).order_by('bone_id').all()
    if dbbones:
        bones_elt = etree.SubElement(root_elt, 'bones')
        for dbbone in dbbones:
            bones_elt.append(dbbone.db2xml())



# pylint: disable = import-outside-toplevel
"""Tests of ``includes.modules`` functions."""

from collections import OrderedDict

from transaction import manager

from ....tests import DBUnitTestCase
# pylint: disable = unused-import
from ....models.dbgroup import DBGroup  # noqa
# pylint: enable = unused-import


# =============================================================================
class UIncludesModulesIncludeme(DBUnitTestCase):
    """Unit test class for :func:`includes.modules.includeme`."""

    # -------------------------------------------------------------------------
    def test_configurator(self):
        """[u:includes.modules.includeme] with a configurator"""
        from ....menu import MENU_ADMIN
        from ....modes import MODE_HOME, MODE_ADMIN
        from ....tests import DummyRootFactory
        from .. import PRINCIPALS_MODULES, SUBMENU_MODULES, includeme
        from ..models import DBModule

        self.request.dbsession.add(
            DBModule(module_id='skeleton', inactive=True))
        manager.commit()
        self.configurator.set_root_factory(DummyRootFactory)
        self.configurator.registry['menu'] = [MENU_ADMIN]

        # With menu
        includeme(self.configurator)
        self.assertIn(
            PRINCIPALS_MODULES[0], self.configurator.registry['principals'])
        self.assertIn(
            SUBMENU_MODULES, self.configurator.registry['menu'][0][4])
        self.assertIn('modules', self.configurator.registry)
        self.assertIn('modules_off', self.configurator.registry)

        # With modes
        self.configurator.registry['modes'] = [MODE_HOME, MODE_ADMIN]
        includeme(self.configurator)
        mode = dict(self.configurator.registry['modes']).get('admin')
        self.assertIsNotNone(mode)
        self.assertEqual(mode[4][4][1], SUBMENU_MODULES)

        # pylint: disable = no-member
        DBModule.__table__.drop()
        # pylint: enable = no-member
        includeme(self.configurator)

    # -------------------------------------------------------------------------
    def test_dictionary(self):
        """[u:includes.modules.includeme] with a dictionary"""
        from .. import includeme

        module_set = set()
        includeme(module_set)
        self.assertNotIn('chrysalio.modules', module_set)

        module_dict = OrderedDict()
        includeme(module_dict)
        self.assertFalse(module_dict)



# -*- coding: utf-8 -*-
"""SQLAlchemy-powered model definitions for modules."""

from pkg_resources import get_distribution

from sqlalchemy import Column, String, Boolean
from webhelpers2.html import literal
import colander

from ...lib.i18n import _
from ...includes.themes import theme_static_prefix
from ...models import DBDeclarativeClass, MODULE_LEN


MODULES_SUFFIX = 'ciomod'


# =============================================================================
class DBModule(DBDeclarativeClass):
    """SQLAlchemy-powered module class."""

    __tablename__ = 'modules'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    module_id = Column(String(MODULE_LEN), primary_key=True)
    inactive = Column(Boolean(name='inactive'), default=True)

    suffix = MODULES_SUFFIX

    # -------------------------------------------------------------------------
    @classmethod
    def table4view(cls, request):
        """Generate the modules table.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if 'modules' not in request.registry or \
           not request.registry['modules']:
            return translate(_('No module.'))
        names = {k: translate(request.registry['modules'][k].name)
                 for k in request.registry['modules']}

        html = '<table>\n<thead>\n'\
            '<tr><th>{active}</th><th>{name}</th><th>{version}</th>'\
            '<th>{depends}</th><th class="cioActions">{actions}</th></tr>\n'\
            '</thead>\n<tbody>\n'.format(
                active=translate(_('Active')), name=translate(_('Name')),
                version=translate(_('Version')),
                depends=translate(_('Dependencies')),
                actions=translate(_('Actions')))
        for module_id in request.registry['modules']:
            module = request.registry['modules'][module_id]
            dependencies = ', '.join(
                [names[k] for k in module.dependencies if k in names]) \
                if module.dependencies else ''
            actions = module.configuration_route(request) or ''
            if actions:
                actions = '<a href="{0}">'\
                    '<img src="{1}{2}" alt="configuration"/></a>'.format(
                        actions, theme_static_prefix(request),
                        '/images/action_configure.png')
            html += '<tr><td>{active}</td><td title="{id}">{name}</td>'\
                '<td>{version}</td><td>{depends}</td>'\
                '<td class="cioActions">{actions}</td></tr>\n'.format(
                    active='' if module_id in request.registry['modules_off']
                    else '✔', id=module_id, name=translate(module.name),
                    version=get_distribution(module_id.split('.')[0]).version,
                    depends=dependencies, actions=actions)
        html += '</tbody>\n</table>\n'

        return literal(html)

    # -------------------------------------------------------------------------
    @classmethod
    def settings_schema(cls, request):
        """Return a Colander schema to edit modules.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: tuple
        :return:
            A tuple such as ``(schema, defaults)``.
        """
        schema = colander.SchemaNode(colander.Mapping())
        for module_id in request.registry['modules']:
            schema.add(colander.SchemaNode(
                colander.Boolean(), name=module_id, missing=False))

        # Defaults
        defaults = {}
        for module_id in request.registry['modules']:
            if module_id not in request.registry['modules_off']:
                defaults[module_id] = True

        return schema, defaults

    # -------------------------------------------------------------------------
    @classmethod
    def table4edit(cls, request, form):
        """Generate the modules table for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if 'modules' not in request.registry or \
           not request.registry['modules']:
            return translate(_('No module.'))
        names = {k: translate(request.registry['modules'][k].name)
                 for k in request.registry['modules']}

        html = '<table>\n<thead>\n'\
            '<tr><th>{selected}</th><th>{name}</th><th>{version}</th>'\
            '<th>{depends}</th></tr>\n</thead>\n<tbody>\n'.format(
                selected=translate(_('Active')), name=translate(_('Name')),
                version=translate(_('Version')),
                depends=translate(_('Dependencies')))
        for module_id in request.registry['modules']:
            module = request.registry['modules'][module_id]
            dependencies = ', '.join(
                [names[k] for k in module.dependencies if k in names]) \
                if module.dependencies else ''
            html += '<tr><td>{selected}</td>'\
                '<td><label for="{cid}" title="{id}">{name}</label></td>'\
                '<td>{version}</td><td>{depends}</td></tr>\n'.format(
                    selected=form.custom_checkbox(module_id),
                    name=translate(module.name),
                    cid=module_id.replace('.', ''), id=module_id,
                    version=get_distribution(module_id.split('.')[0]).version,
                    depends=dependencies)
        html += '</tbody>\n</table>\n'

        return literal(html)



"""Widget to create tabs."""

from webhelpers2.html import literal


# =============================================================================
class Tabset(object):
    """A class to manage tabs.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str tabset_id:
        ID of the tab set.
    :param list labels:
        A list of translation strings.
    """

    # -------------------------------------------------------------------------
    def __init__(self, request, tabset_id, labels):
        """Constructor method."""
        self.request = request
        self.tabset_id = tabset_id
        self.tabs = [Tab(self, i, k) for i, k in enumerate(labels)]

    # -------------------------------------------------------------------------
    def begin(self):
        """Output the opening tag of the ``TabSet`` and its table of content as
        an ``<ul>`` structure.

        :rtype: webhelpers2.html.literal
        """
        translate = self.request.localizer.translate
        html = '<div id="{0}" class="cioTabset">\n'\
            '<ul class="cioTabsetToc">\n'.format(self.tabset_id)
        for tab in self.tabs:
            html += '  <li id="{tabset_id}{index}">'\
                '<a class="cioTab" href="#{tabset_id}{index}_">'\
                '<span>{label}</span></a></li>\n'.format(
                    tabset_id=self.tabset_id, index=tab.index,
                    label=translate(tab.label))
        html += '</ul>\n<div class="cioClear"></div>\n'
        return literal(html)

    # -------------------------------------------------------------------------
    @classmethod
    def end(cls):
        """End a tab set.

        :rtype: webhelpers2.html.literal
        """
        return literal('</div>\n')


# =============================================================================
class Tab(object):
    """A class to manage one tab.

    :type  tabset: Tabset
    :param tabset:
        Parent tab set.
    :param int index:
        Index of the tab.
    :param str label:
        Label of the tab.
    """

    # -------------------------------------------------------------------------
    def __init__(self, tabset, index, label):
        """Constructor method."""
        self._tabset = tabset
        self.index = index
        self.label = label

    # -------------------------------------------------------------------------
    def begin(self):
        """Open a tab zone.

        :rtype: webhelpers2.html.literal
        :return:
            Opening ``fieldset`` structure with legend.
        """
        return literal(
            '<fieldset class="cioTabContent" id="{tabset_id}{index}_">\n'
            '  <legend><span>{label}</span></legend>\n'.format(
                tabset_id=self._tabset.tabset_id,
                index=self.index,
                label=self._tabset.request.localizer.translate(self.label)))

    # -------------------------------------------------------------------------
    @classmethod
    def end(cls):
        """Close a tab zone.

        :rtype: webhelpers2.html.literal
        :return:
            Closing ``fieldset`` structure.
        """
        return literal('</fieldset>\n')



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.modes`` class methods."""

from pyramid.httpexceptions import HTTPNotFound, HTTPForbidden

from . import ConfiguratorTestCase


MENU_ADMIN = (
    '{theme}/images/menu_admin.png', 'Administration', None, None,
    ((None, 'Profile Management', 'profile-create', 'profile_index', None),
     (None, 'User Management', None, 'user_index', None)))
MODE_HOME = ('home', (
    '{theme}/images/menu_home.png', 'Home', None, 'home', None))
MODE_TEST = ('test', (
    '{theme}/images/menu_test1.png', 'Test', None, 'home', None))
MODE_ADMIN = ('admin', (
    '{theme}/images/menu_admin.png', ('Administration'), 'mode-admin',
    'settings_view', MENU_ADMIN))


# =============================================================================
class ULibModesModes(ConfiguratorTestCase):
    """Unit test class for :class:`lib.modes.Modes`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from pyramid.testing import DummyRequest
        from ..models.dbsettings import DBSettings

        super(ULibModesModes, self).setUp()
        self.configurator.add_route('home', '/')
        self.configurator.add_route('mode', '/mode/{mode_id}')
        self.configurator.include('..security')

        self.request = DummyRequest()
        self.request.registry['settings'] = DBSettings.settings_defaults.copy()

    # -------------------------------------------------------------------------
    def test_label(self):
        """[u:lib.modes.Modes.label]"""
        from ..lib.modes import Modes

        modes = Modes(self.request, 'modes', [MODE_HOME, MODE_ADMIN])
        self.assertEqual(modes.label(), 'Home')
        self.assertIn(modes.uid, self.request.session)
        self.assertEqual(self.request.session[modes.uid][0], 'home')

    # -------------------------------------------------------------------------
    def test_menu_framework(self):
        """[u:lib.modes.Modes.menu_framework]"""
        from ..lib.modes import Modes

        # New activation
        modes = Modes(self.request, 'modes', [MODE_ADMIN, MODE_TEST])
        framework = modes.menu_framework()
        self.assertIsInstance(framework, tuple)
        self.assertEqual(len(framework), 2)
        self.assertEqual(framework[0], MENU_ADMIN[4][0])

        # Unknown activated mode
        modes = Modes(self.request, 'modes', [MODE_HOME, MODE_ADMIN])
        self.request.session[modes.uid] = [
            'foo', 'Foo', [MODE_HOME, MODE_ADMIN]]
        framework = modes.menu_framework()
        self.assertIsInstance(framework, tuple)
        self.assertEqual(len(framework), 0)

        # Kwown activated mode
        self.request.session[modes.uid] = [
            'admin', 'Administration', [MODE_HOME, MODE_ADMIN]]
        framework = modes.menu_framework()
        self.assertIsInstance(framework, tuple)
        self.assertEqual(len(framework), 2)

    # -------------------------------------------------------------------------
    def test_select(self):
        """[u:lib.modes.Modes.select]"""
        from ..lib.modes import Modes

        # Unknown mode
        modes = Modes(self.request, 'modes', [MODE_HOME, MODE_ADMIN])
        self.assertRaises(HTTPNotFound, modes.select, 'foo')

        # Forbidden mode
        self.assertRaises(HTTPForbidden, modes.select, 'admin')

        # Authorized mode
        route = modes.select('home')
        self.assertEqual(route, '/')

    # -------------------------------------------------------------------------
    def test_xhtml(self):
        """[u:lib.modes.Modes.xhtml]"""
        from ..lib.modes import Modes

        # 1 authorized mode
        modes = Modes(self.request, 'modes', [MODE_HOME, MODE_ADMIN])
        html = modes.xhtml()
        self.assertFalse(html)
        self.assertIn(modes.uid, self.request.session)
        self.assertEqual(self.request.session[modes.uid][0], 'home')
        self.assertEqual(self.request.session[modes.uid][1], 'Home')
        self.assertFalse(self.request.session[modes.uid][2])

        # 2 authorized modes
        modes = Modes(
            self.request, 'modes', [MODE_HOME, MODE_TEST, MODE_ADMIN])
        self.request.session[modes.uid] = ['home', 'Home', None]
        html = modes.xhtml()
        self.assertIn('cioModes-{0}'.format(modes.uid), html)

        # Cache
        html = modes.xhtml()
        self.assertIn('cioModes-{0}'.format(modes.uid), html)

        # 0 authorized mode
        modes = Modes(self.request, 'modes', [MODE_ADMIN])
        self.request.session[modes.uid] = ['home', 'Home', None]
        html = modes.xhtml()
        self.assertFalse(html)

    # -------------------------------------------------------------------------
    def test_invalidate(self):
        """[u:lib.modes.Modes.invalidate]"""
        from ..lib.modes import Modes

        self.request.session['modes'] = ['home', 'Home', '']
        Modes.invalidate(self.request, 'modes')
        self.assertIn('modes', self.request.session)
        self.assertEqual(self.request.session['modes'][0], 'home')
        self.assertEqual(self.request.session['modes'][1], 'Home')
        self.assertIsNone(self.request.session['modes'][2])



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.utils`` functions."""

from os import walk
from os.path import join, basename, dirname, exists, relpath
from shutil import rmtree, copy
from sys import version_info
from base64 import b64encode
from datetime import date, datetime, timedelta
from filecmp import cmpfiles
from re import compile as re_compile

from unittest import TestCase

from . import TmpDirTestCase


# =============================================================================
class ULibUtilsTostr(TestCase):
    """Unit test class for :func:`lib.utils.tostr`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.tostr]"""
        from ..lib.utils import tostr

        text = tostr('foo')
        self.assertIsInstance(text, str)
        self.assertEqual(text, 'foo')

        text = tostr('foo'.encode('utf8'))
        self.assertIsInstance(text, str)
        self.assertEqual(text, 'foo')

        text = tostr(b'foo')
        self.assertIsInstance(text, str)
        self.assertEqual(text, 'foo')

        text = tostr(bytes('foo'.encode('utf8')))
        self.assertIsInstance(text, str)
        self.assertEqual(text, 'foo')

        text = tostr('é Ô')
        self.assertIsInstance(text, str)
        self.assertEqual(
            text, version_info > (3, 0) and 'é Ô' or 'é Ô'.encode('utf8'))


# =============================================================================
class ULibUtilsToUnicode(TestCase):
    """Unit test class for :func:`lib.utils.tounicode`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.unicode]"""
        from ..lib.utils import tounicode

        text = tounicode('foo')
        self.assertIsInstance(text, type(''))
        self.assertEqual(text, 'foo')

        text = tounicode('foo'.encode('utf8'))
        self.assertIsInstance(text, type(''))
        self.assertEqual(text, 'foo')

        text = tounicode(b'foo')
        self.assertIsInstance(text, type(''))
        self.assertEqual(text, 'foo')

        text = tounicode(bytes('foo'.encode('utf8')))
        self.assertIsInstance(text, type(''))
        self.assertEqual(text, 'foo')

        text = tounicode('é Ô')
        self.assertIsInstance(text, type(''))
        self.assertEqual(text, 'é Ô')


# =============================================================================
class ULibUtilsLoadGuessingEncoding(TestCase):
    """Unit test class for :func:`lib.utils.laoad_guessing_encoding`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.unicode]"""
        from ..lib.utils import load_guessing_encoding
        from . import TEST_LATIN1, TEST_UTF8

        self.assertIsNotNone(load_guessing_encoding(TEST_UTF8))
        self.assertIsNotNone(load_guessing_encoding(TEST_LATIN1))


# =============================================================================
class ULibUtilsCopyContent(TestCase):
    """Unit test class for :func:`lib.utils.copy_content`."""

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from . import TEST_DIR

        super(ULibUtilsCopyContent, self).tearDown()
        if exists(TEST_DIR):
            rmtree(TEST_DIR)

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.copy_content]"""
        from ..lib.utils import EXCLUDED_FILES, copy_content
        from . import TEST_DIR

        # With specific excluded files
        exclude = EXCLUDED_FILES + ('hugo.txt',)
        source_dir = join(dirname(__file__), 'Texts')
        copy_content(source_dir, TEST_DIR, exclude)
        for path, dirs, files in walk(source_dir):
            for name in dirs + files:
                copy_file = join(TEST_DIR, relpath(path, source_dir), name)
                if name in exclude:
                    self.assertFalse(exists(copy_file))
                else:
                    self.assertTrue(exists(copy_file))

        # With normal excluded files
        copy_content(source_dir, TEST_DIR)
        self.assertTrue(exists(join(TEST_DIR, 'Écrits', 'XIXe', 'hugo.txt')))


# =============================================================================
class ULibUtilsCopyContentRe(TestCase):
    """Unit test class for :func:`lib.utils.copy_content_re`."""

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from . import TEST_DIR

        super(ULibUtilsCopyContentRe, self).tearDown()
        if exists(TEST_DIR):
            rmtree(TEST_DIR)

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.copy_content]"""
        from ..lib.utils import copy_content_re
        from . import TEST_DIR

        source_dir = join(dirname(__file__), 'Texts')
        exclude = re_compile('(^X|1.rst$)')
        copy_content_re(source_dir, TEST_DIR, exclude)
        self.assertTrue(exists(join(TEST_DIR, 'utf8.txt')))
        self.assertFalse(exists(join(TEST_DIR, 'latin1.rst')))
        self.assertTrue(exists(join(TEST_DIR, 'Écrits')))
        self.assertFalse(exists(join(TEST_DIR, 'Écrits', 'XXe')))
        self.assertFalse(exists(join(TEST_DIR, 'Écrits', 'XIXe')))


# =============================================================================
class ULibUtilsMakeId(TestCase):
    """Unit test class for :func:`lib.utils.make_id`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.make_id]"""
        from ..lib.utils import make_id

        name = '12Test___Té*t.;?!'
        self.assertEqual(make_id(''), '')
        self.assertEqual(make_id(name), '12test___té*t.;?!')
        self.assertEqual(make_id(name, 'standard'), '12Test_Té_t._')
        self.assertEqual(make_id(name, 'token'), '12test_te_t._')
        self.assertEqual(make_id(name, 'token', 6), '12test')
        self.assertEqual(make_id(name, 'xmlid'), '_12test_te_t._')
        self.assertEqual(make_id(name, 'class'), '12Test_Te_t_')
        self.assertEqual(make_id(name, 'no_accent'), '12Test___Te*t.;?!')


# =============================================================================
class ULibUtilsMakeDigest(TestCase):
    """Unit test class for :func:`lib.utils.make_digest`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.make_digest]"""
        from ..lib.utils import make_digest, tounicode

        digest = make_digest(
            '/browse/directory/Sandbox/Écrits/XXe/de Saint Exupéry.txt')
        self.assertEqual(digest, '1d4fe4892722b034801a7bac95fd3c33')
        digest = make_digest(tounicode(
            '/browse/directory/Sandbox/Écrits/XIXe/hugo.txt'.encode('utf8')))
        self.assertEqual(digest, '34a21e0695daed8a351a54b0116c3f70')


# =============================================================================
class ULibUtilsNormalizeSpaces(TestCase):
    """Unit test class for :func:`lib.utils.normalize_spaces`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.normalize_spaces]"""
        from ..lib.utils import normalize_spaces

        text = 'test test1       é  _    '
        self.assertEqual(normalize_spaces(None), None)
        self.assertEqual(normalize_spaces('test  test '), 'test test')
        self.assertEqual(normalize_spaces(text), 'test test1 é _')
        self.assertEqual(normalize_spaces(text, 9), 'test test')


# =============================================================================
class ULibUtilsCamelCase(TestCase):
    """Unit test class for :func:`lib.utils.camel_case`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.camel_case]"""
        from ..lib.utils import camel_case

        self.assertEqual(camel_case('xml2html'), 'Xml2Html')
        self.assertEqual(camel_case('LaTeX'), 'LaTeX')
        self.assertEqual(camel_case('laTeX'), 'LaTeX')
        self.assertEqual(camel_case('my_way'), 'MyWay')
        self.assertEqual(camel_case('my way'), 'MyWay')
        self.assertEqual(camel_case('my-way'), 'My-Way')


# =============================================================================
class ULibUtilsShorten(TestCase):
    """Unit test class for :func:`lib.utils.shorten`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.shortent]"""
        from ..lib.utils import shorten

        self.assertEqual(shorten('12 45  7890', 10), '12 45 7890')
        self.assertEqual(shorten('12 45  789 12', 10), '12 45 789…')
        self.assertEqual(shorten('12 45  789012', 10, '...'), '12 45 7...')
        self.assertEqual(shorten('12 45  789012', 10, ''), '12 45 7890')


# =============================================================================
class ULibUtilsEncryptDecrypt(TestCase):
    """Unit test class for :func:`lib.utils.encrypt` and
    :func:`lib.utils.decrypt`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.encrypt/decrypt]"""
        from ..lib.utils import encrypt, decrypt

        self.assertIsNone(encrypt(None, 'seekrit'))
        self.assertEqual(encrypt('', 'seekrit'), '')
        self.assertIsNone(decrypt(None, 'seekrit'))
        self.assertEqual(decrypt('', 'seekrit'), '')

        crypted = encrypt('protectme', 'seekrit')
        self.assertEqual(len(crypted), 44)
        self.assertEqual(decrypt(crypted, 'seekrit'), 'protectme')

        crypted = encrypt(b'protectme', 'seekrit')
        self.assertEqual(decrypt(crypted, 'seekrit'), 'protectme')

        crypted = encrypt(bytes('protège moi'.encode('utf8')), 'seekrit')
        self.assertEqual(len(crypted), 44)
        self.assertEqual(decrypt(crypted, 'seekrit'), 'protège moi')

        self.assertIsNone(decrypt('foo', 'seekrit'))
        self.assertIsNone(decrypt(b64encode(b'foo'), 'seekrit'))


# =============================================================================
class ULibUtilsToken(TestCase):
    """Unit test class for :func:`lib.utils.token`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.token]"""
        from ..lib.utils import token

        result = token()
        self.assertTrue(len(result) > 7)
        self.assertTrue(len(result) < 17)

        result = token(32)
        self.assertEqual(len(result), 32)


# =============================================================================
class ULibUtilsExecute(TestCase):
    """Unit test class for :func:`lib.utils.execute`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.execute]"""
        from ..lib.utils import execute

        self.assertEqual(
            execute(['echo', 'test'], None, False), ('test', ''))
        self.assertEqual(
            execute(['nice', 'echo', 'test'], None, False), ('test', ''))
        self.assertEqual(
            execute(['echo', 'testé'], None, False), ('testé', ''))
        self.assertEqual(
            execute(['echo', '\xa9'], None, False), ('\xa9', ''))
        self.assertEqual(
            execute(['echo', 'test1', 'test2'], None, False),
            ('test1 test2', ''))

        self.assertEqual(
            execute(['echo2'], None, False), ('', '"${c}" failed: ${e}'))
        self.assertEqual(
            execute(['echo', 'test'], 'cwd', True),
            ('', '"${c}" failed: ${e}'))
        self.assertEqual(
            execute(['echo', 'test'], None, True), ('test', '"${c}" failed'))


# =============================================================================
class ULibUtilsFullUrl(TestCase):
    """Unit test class for :func:`lib.utils.full_url`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.full_url]"""
        from ..lib.utils import full_url

        url = 'http://hg.publiforge.org'
        new_url = full_url(url)
        self.assertEqual(new_url, url)

        new_url = full_url(url, 'foo')
        self.assertEqual(new_url, 'http://foo@hg.publiforge.org')

        new_url = full_url(url, 'foo', 'bar')
        self.assertEqual(new_url, 'http://foo:bar@hg.publiforge.org')


# =============================================================================
class ULibUtilsMimetypeGet(TestCase):
    """Unit test class for :func:`lib.utils.mimetype_get`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.mimetype_get]"""
        from . import TEST_XXX, TEST_CSS
        from ..lib.utils import mimetype_get

        # A directory
        mimetype, file_type = mimetype_get(dirname(__file__))
        self.assertEqual(mimetype, 'directory')
        self.assertEqual(file_type, 'directory')

        # An unknown file
        mimetype, file_type = mimetype_get(TEST_XXX)
        self.assertEqual(mimetype, 'unknown')
        self.assertEqual(file_type, 'unknown')

        # A CSS file
        mimetype, file_type = mimetype_get(TEST_CSS)
        self.assertEqual(mimetype, 'text/css')
        self.assertEqual(file_type, 'css')


# =============================================================================
class ULibUtilsDeltatimeLabel(TestCase):
    """Unit test class for :func:`lib.utils.deltatime_label`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.deltatime_label]"""
        from ..lib.utils import deltatime_label

        self.assertEqual(deltatime_label(lang='en'), '0 second')
        self.assertEqual(deltatime_label(days=1, lang='en'), '1 day')
        self.assertEqual(deltatime_label(days=2, lang='en'), '2 days')
        self.assertEqual(deltatime_label(hours=1, lang='en'), '1 hour')
        self.assertEqual(deltatime_label(hours=2, lang='en'), '2 hours')
        self.assertEqual(deltatime_label(minutes=1, lang='en'), '1 minute')
        self.assertEqual(deltatime_label(minutes=2, lang='en'), '2 minutes')
        self.assertEqual(deltatime_label(seconds=1, lang='en'), '1 second')
        self.assertEqual(deltatime_label(seconds=2, lang='en'), '2 seconds')

        self.assertEqual(
            deltatime_label(seconds=303, lang='en'), '5 minutes 3 seconds')

        self.assertEqual(
            deltatime_label(seconds=1, hours=25, days=1, lang='en'),
            '2 days 1 hour 1 second')


# =============================================================================
class ULibUtilsAge(TestCase):
    """Unit test class for :func:`lib.utils.age`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.age]"""
        from ..lib.utils import age

        self.assertEqual(age(None), '')

        mtime = datetime.now() + timedelta(seconds=12)
        self.assertEqual(age(mtime), '0 second')
        mtime = datetime.now() - timedelta(seconds=1)
        self.assertEqual(age(mtime), '1 second')
        mtime = datetime.now() - timedelta(seconds=12)
        self.assertIn('seconds', age(mtime))

        mtime = datetime.now() - timedelta(minutes=1)
        self.assertEqual(age(mtime), '1 minute')
        mtime = datetime.now() - timedelta(minutes=12)
        self.assertIn('minutes', age(mtime))

        mtime = datetime.now() - timedelta(hours=1)
        self.assertEqual(age(mtime), '1 hour')
        mtime = datetime.now() - timedelta(hours=12)
        self.assertIn('hours', age(mtime))

        mtime = datetime.now() - timedelta(days=1)
        self.assertEqual(age(mtime), '1 day')
        mtime = datetime.now() - timedelta(days=3)
        self.assertIn('days', age(mtime))

        mtime = datetime.now() - timedelta(weeks=1)
        self.assertEqual(age(mtime), '1 week')
        mtime = datetime.now() - timedelta(weeks=3)
        self.assertIn('weeks', age(mtime))

        mtime = datetime.now() - timedelta(weeks=12)
        self.assertIn('months', age(mtime))

        mtime = datetime.now() - timedelta(weeks=60)
        self.assertIn('-', age(mtime))


# =============================================================================
class ULibUtilsSizeLabel(TestCase):
    """Unit test class for :func:`lib.utils.size_label`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.size_label]"""
        from ..lib.i18n import _, translate
        from ..lib.utils import size_label

        # Directory, 1 element
        self.assertEqual(
            translate(size_label(1, True)),
            translate(_('${n} item', {'n': 1})))

        # Directory, several elements
        self.assertEqual(
            translate(size_label(10, True)),
            translate(_('${n} items', {'n': 10})))

        # File, bytes
        self.assertEqual(size_label(512), '512 o')

        # File, Kio
        self.assertEqual(size_label(2048), '2.0 Kio')

        # File, Mio
        self.assertEqual(size_label(3250590), '3.1 Mio')

        # File, Gio
        self.assertEqual(size_label(4831838208), '4.5 Gio')


# =============================================================================
class ULibUtilsConvertValue(TestCase):
    """Unit test class for :func:`lib.utils.convert_value`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.convert_value]"""
        from ..lib.utils import convert_value

        # Boolean
        self.assertIsInstance(convert_value('boolean', 'true'), bool)
        self.assertTrue(convert_value('boolean', 'foo'))
        self.assertFalse(convert_value('boolean', 'false'))
        self.assertFalse(convert_value('boolean', ''))

        # Integer
        self.assertIsInstance(convert_value('integer', '12'), int)
        self.assertEqual(convert_value('integer', '12'), 12)
        self.assertEqual(convert_value('integer', 'foo'), 0)

        # Float
        self.assertIsInstance(convert_value('decimal', '12'), float)
        self.assertIsInstance(convert_value('decimal', '3.14'), float)
        self.assertEqual(convert_value('decimal', '12'), 12.0)
        self.assertEqual(convert_value('decimal', '3.14'), 3.14)
        self.assertEqual(convert_value('decimal', 'foo'), 0.0)

        # Date/time
        self.assertIsNone(convert_value('datetime', 'foo'))
        self.assertIsInstance(
            convert_value('datetime', '2000-12-10T10:20:00'), datetime)
        self.assertEqual(
            convert_value('datetime', '1997-10-24T09:45:12'),
            datetime(1997, 10, 24, 9, 45, 12))

        # Date
        self.assertIsNone(convert_value('date', 'foo'))
        self.assertIsInstance(convert_value('date', '2000-12-10'), date)
        self.assertEqual(
            convert_value('date', '1997-10-24'), date(1997, 10, 24))

        # String
        self.assertIsInstance(convert_value('string', 'foo'), str)
        self.assertEqual(convert_value('string', 'foo'), 'foo')


# =============================================================================
class ULibUtilsCheckChrysalioJs(TmpDirTestCase):
    """Unit test class for :func:`lib.utils.check_chrysalio_js`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.check_chrysalio_js]"""
        from ..lib.utils import check_chrysalio_js
        from . import TEST_DIR

        js_dir = TEST_DIR
        check_chrysalio_js(join(js_dir, 'Foo'))

        cio_js_dir = join(dirname(__file__), '..', 'Static', 'Js')
        copy(join(cio_js_dir, 'jquery.js'), join(js_dir, 'jquery.js'))
        with open(join(js_dir, 'chrysalio.js'), 'w'):
            pass
        js_files = ('jquery.js', 'js.cookie.js', 'chrysalio.js')
        compare = cmpfiles(cio_js_dir, js_dir, js_files)
        self.assertEqual(compare[0], ['jquery.js'])
        self.assertEqual(compare[1], ['chrysalio.js'])
        self.assertEqual(len(compare[2]), 1)

        check_chrysalio_js(js_dir)

        compare = cmpfiles(cio_js_dir, js_dir, js_files)
        self.assertEqual(len(compare[0]), 2)
        self.assertEqual(len(compare[1]), 0)
        self.assertEqual(len(compare[2]), 1)


# =============================================================================
class ULibUtilsCheckChrysalioCss(TmpDirTestCase):
    """Unit test class for :func:`lib.utils.check_chrysalio_css`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.check_chrysalio_js]"""
        from ..lib.utils import check_chrysalio_css
        from . import TEST_DIR

        css_dir = TEST_DIR
        check_chrysalio_css(join(css_dir, 'Foo'))

        cio_css_dir = join(dirname(__file__), '..', 'Static', 'Css')
        copy(join(cio_css_dir, 'jquery-ui.css'),
             join(css_dir, 'jquery-ui.css'))
        with open(join(css_dir, 'jquery-ui.css'), 'w'):
            pass
        css_files = ('jquery.css', 'jquery-ui.css')
        compare = cmpfiles(cio_css_dir, css_dir, css_files)
        self.assertEqual(len(compare[0]), 0)
        self.assertEqual(compare[1], ['jquery-ui.css'])
        self.assertEqual(len(compare[2]), 1)

        check_chrysalio_css(css_dir)

        compare = cmpfiles(cio_css_dir, css_dir, css_files)
        self.assertEqual(len(compare[0]), 1)
        self.assertEqual(len(compare[1]), 0)
        self.assertEqual(len(compare[2]), 1)


# =============================================================================
class ULibUtilsCommonDirectory(TmpDirTestCase):
    """Unit test class for :func:`lib.utils.common_directory`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.common_directory]"""
        from . import ATTACHMENTS_DIR, TEST1_SVG, TEST5_SVG
        from ..lib.utils import common_directory

        self.assertIsNone(common_directory([]))
        self.assertIsNone(common_directory(
            [join(ATTACHMENTS_DIR, 'foo', 'bar')]))

        self.assertEqual(common_directory([TEST1_SVG]), dirname(TEST1_SVG))
        self.assertEqual(
            common_directory([TEST1_SVG, TEST5_SVG]),
            join(ATTACHMENTS_DIR, 'Users'))


# =============================================================================
class ULibUtilsHashAsdminPassword(TmpDirTestCase):
    """Unit test class for :func:`lib.utils.hash_admin_password`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.hash_admin_password]"""
        from . import TEST_DIR, TEST_INI
        from ..lib.utils import hash_admin_password

        copy(TEST_INI, TEST_DIR)

        hash_admin_password(TEST_DIR, (basename(TEST_INI),), 'adminpwd')
        with open(join(TEST_DIR, basename(TEST_INI)), 'r') as hdl:
            content = hdl.read()
        self.assertIn('admin.password = $2b$12$', content)


# =============================================================================
class ULibUtilsRst2Html(TestCase):
    """Unit test class for :func:`lib.utils.rst2html`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.utils.rst2html]"""
        from ..lib.utils import rst2html
        self.assertEqual(
            rst2html(''), '')
        self.assertEqual(
            rst2html(None), None)
        self.assertEqual(
            rst2html('foo_'), 'foo_')
        self.assertEqual(
            rst2html('test'), 'test')
        self.assertEqual(
            rst2html('*emphasis*'), '<em>emphasis</em>')
        self.assertEqual(
            rst2html('**strong**'), '<strong>strong</strong>')
        self.assertEqual(
            rst2html('``inline literal``'),
            '<tt class="docutils literal">inline literal</tt>')



# -*- coding: utf-8 -*-
"""User management view callables."""

from os.path import join, exists
from datetime import date, datetime
from hashlib import sha1
from shutil import rmtree

from sqlalchemy import desc
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError
from colander import Mapping, SchemaNode, String, All, Length, Email

from pyramid.httpexceptions import HTTPFound, HTTPNotFound, HTTPForbidden
from pyramid.view import view_config
from pyramid.security import NO_PERMISSION_REQUIRED

from ..lib.i18n import _
from ..lib.log import log_info
from ..lib.utils import make_id, age
from ..lib.form import get_action, SameAs, Form
from ..lib.paging import PAGE_SIZES, Paging
from ..lib.filter import Filter
from ..lib.tabset import Tabset
from ..lib.mailing import Mailing
from ..lib.attachment import attachment_url, attachment_update
from ..includes.themes import theme_static_prefix
from ..models import LABEL_LEN
from ..models.populate import xml2db, db2web, web2db
from ..models.dbprofile import DBProfile
from ..models.dbuser import DBUser, DBUserProfile
from ..models.dbgroup import DBGroup, DBGroupUser
from . import BaseView


# =============================================================================
class UserView(BaseView):
    """Class to manage user views.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    _DBUser = DBUser
    _xml2db = (xml2db,)
    _invitation_html = 'chrysalio:Templates/email_invitation.pt'
    _invitation_text = 'chrysalio:Templates/email_invitation.txt'
    _reset_password_html = 'chrysalio:Templates/email_reset_password.pt'
    _reset_password_text = 'chrysalio:Templates/email_reset_password.txt'

    # -------------------------------------------------------------------------
    @view_config(
        route_name='user_index', renderer='chrysalio:Templates/user_index.pt',
        permission='user-view')
    @view_config(route_name='user_index', renderer='json', xhr=True)
    def index(self):
        """List all users."""
        # Ajax
        i_creator = self._request.has_permission('user-create')
        if self._request.is_xhr:
            if i_creator:
                self._import_users()
            return {}

        # Action
        action, items = get_action(self._request)
        i_editor = self._request.has_permission('user-edit')
        if action[:4] == 'mel!' and i_editor:
            errors = self._send_invitation(items)
            if not errors:  # pragma: nocover
                self._request.session.flash(_('The invitation has been sent.'))
            else:
                for error in errors:
                    self._request.session.flash(error, 'alert')
        elif action[:4] == 'del!' and i_creator:
            self._delete_users(items)
        elif action == 'imp!' and i_creator:
            self._import_users()
        elif action[:4] == 'exp!' and i_editor:
            action = self._users2response(items)
            if action:
                return action

        # Filter
        paging_id = 'users'
        pfilter = Filter(
            self._request, paging_id, (
                ('login', _('Login'), False, ''),
                ('last_name', _('Last name'), False, ''),
                ('email', _('Email'), False, ''),
                # ('email_hidden', _('Hidden Email'), False, True),
                ('status', _('Status'), False,
                 [('', ' ')] + list(self._DBUser.status_labels.items()))),
            (('status', '=', 'active'),),
            remove=action[:4] == 'crm!' and action[4:] or None)

        # Paging
        defaults = Paging.params(self._request, paging_id, '+last_name')
        dbquery = pfilter.sql(
            self._request.dbsession.query(self._DBUser), 'users')
        oby = getattr(self._DBUser, defaults['sort'][1:])
        dbquery = dbquery.order_by(
            desc(oby) if defaults['sort'][0] == '-' else oby)
        user_paging = Paging(self._request, paging_id, dbquery, defaults)
        user_paging.set_current_ids('user_id')

        # Form & completed action
        form = Form(self._request, defaults=defaults)
        form.forget('filter_value')
        if action and action[3] == '!':
            action = ''

        # Breadcrumbs
        self._request.breadcrumbs(_('Users'), 2)

        return {
            'age': age, 'action': action, 'items': items, 'form': form,
            'pfilter': pfilter, 'paging': user_paging,
            'status_labels': self._DBUser.status_labels,
            'PAGE_SIZES': PAGE_SIZES, 'i_creator': i_creator,
            'i_editor': i_editor, 'has_attachments': bool(
                self._request.registry.settings.get('attachments')),
            'download_max_size': self._request.registry['settings'][
                'download-max-size'],
            'attachment_url': attachment_url}

    # -------------------------------------------------------------------------
    @view_config(route_name='user_index_filter', renderer='json', xhr=True)
    def index_filter(self):
        """Return a list to autocomplete a filter field."""
        return Filter.sql_autocomplete(self._request, self._DBUser)

    # -------------------------------------------------------------------------
    @view_config(
        route_name='user_view', renderer='chrysalio:Templates/user_view.pt')
    @view_config(
        route_name='user_account', renderer='chrysalio:Templates/user_view.pt')
    def view(self):
        """Show user settings."""
        if not self.can_view():
            raise HTTPForbidden()

        dbuser = self._get_user(str(self._request.session['user']['user_id']))
        is_me = dbuser.user_id == self._request.session['user']['user_id']
        picture = self._request.registry.settings.get('attachments') and (
            attachment_url(
                self._request, dbuser.attachments_dir, dbuser.attachments_key,
                dbuser.picture) or
            (dbuser.honorific == 'Mrs' and
             '{0}/images/user_picture_girl.svg'.format(
                 theme_static_prefix(self._request))) or
            '{0}/images/user_picture_boy.svg'.format(
                theme_static_prefix(self._request)))
        navigator = Paging.navigator(
            self._request, 'users', dbuser.user_id,
            self._request.route_path('user_view', user_id='_ID_')) \
            if self._request.matched_route.name == 'user_view' else ''

        action = get_action(self._request)[0]
        if action == 'exp!':
            action = self._users2response((dbuser.user_id,))
            if action:
                return action
        elif action == 'mel!':
            errors = self._send_invitation((dbuser.user_id,))
            if not errors:  # pragma: nocover
                self._request.session.flash(_('The invitation has been sent.'))
            else:
                self._request.session.flash(errors[0], 'alert')
            action = ''

        # Breadcrumbs
        user_name = '{0} {1}'.format(
            dbuser.first_name or '', dbuser.last_name).strip()
        if self._request.referrer and  \
           '/user/view/' in self._request.referrer:
            self._request.breadcrumbs.pop()
        self._request.breadcrumbs(
            _('My Account') if is_me else _('${n} Account', {'n': user_name}),
            replace=self._request.route_path(
                'user_edit', user_id=dbuser.user_id))
        return {
            'form': Form(self._request),
            'tabset': Tabset(
                self._request, 'tabUser', dbuser.settings_tabs(self._request)),
            'dbuser': dbuser, 'is_me': is_me, 'user_name': user_name,
            'navigator': navigator, 'picture': picture,
            'i_editor': self._request.has_permission('user-edit')}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='user_create', renderer='chrysalio:Templates/user_edit.pt')
    @view_config(
        route_name='user_edit', renderer='chrysalio:Templates/user_edit.pt')
    @view_config(route_name='user_edit', renderer='json', xhr=True)
    def edit(self):
        """Create or edit user settings."""
        # Authorization
        dbuser = self._get_user() if 'user_id' in self._request.matchdict \
            else None
        if (not dbuser and not self.can_create()) or \
           (dbuser and not self.can_edit()):
            raise HTTPForbidden()

        # Ajax
        if self._request.is_xhr:
            if dbuser is not None:
                dbuser.attachments_key, dbuser.picture = attachment_update(
                    self._request, dbuser.attachments_dir,
                    dbuser.attachments_key, self._request.POST['picture'],
                    replace=dbuser.picture,
                    prefix=dbuser.login.replace('@', '_')[:12])
                dbuser.account_update = datetime.now()
                log_info(self._request, 'user_update_picture', dbuser.login)
            return {}

        # Initialization
        is_me = dbuser and \
            dbuser.user_id == self._request.session['user']['user_id']
        profiles = {
            k.profile_id: (
                k.label(self._request), k.description(self._request))
            for k in self._request.dbsession.query(DBProfile)}
        groups = {
            k.group_id: (
                k.label(self._request), k.description(self._request))
            for k in self._request.dbsession.query(DBGroup)}

        # Form and action
        form = Form(
            self._request,
            *self._DBUser.settings_schema(
                self._request, profiles, groups, dbuser),
            obj=dbuser)
        action = get_action(self._request)[0]
        if action == 'pct!' and dbuser is not None:
            dbuser.attachments_key, dbuser.picture = attachment_update(
                self._request, dbuser.attachments_dir, dbuser.attachments_key,
                self._request.POST['picture'], replace=dbuser.picture,
                prefix=dbuser.login.replace('@', '_')[:12])
            dbuser.account_update = datetime.now()
            log_info(self._request, 'user_update_picture', dbuser.login)
        elif action == 'sav!' and form.validate():
            dbuser = self._save(dbuser, profiles, groups, form.values)
            if dbuser is not None:
                if is_me:
                    dbuser.set_session(self._request)
                if 'user_id' not in self._request.matchdict:
                    self._request.breadcrumbs.pop()
                log_info(
                    self._request, 'user_id' in self._request.matchdict and
                    'user_edit' or 'user_create', dbuser.login)
                return HTTPFound(self._request.route_path(
                    'user_view', user_id=dbuser.user_id))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Picture
        picture = \
            dbuser and self._request.registry.settings.get('attachments') and (
                attachment_url(
                    self._request, DBUser.attachments_dir,
                    dbuser.attachments_key, dbuser.picture) or
                (dbuser.honorific == 'Mrs' and
                 '{0}/images/user_picture_girl.svg'.format(
                     theme_static_prefix(self._request))) or
                '{0}/images/user_picture_boy.svg'.format(
                    theme_static_prefix(self._request)))

        # Breadcrumbs
        user_name = dbuser and '{0} {1}'.format(
            dbuser.first_name or '', dbuser.last_name).strip()
        if self._request.referrer and  \
           '/user/account' in self._request.referrer:
            self._request.breadcrumbs.pop()
        if not dbuser:
            self._request.breadcrumbs(_('User Creation'))
        else:
            self._request.breadcrumbs(
                _('My Account Edition') if is_me else _(
                    '${n} Edition', {'n': user_name}),
                replace=self._request.route_path(
                    'user_view', user_id=dbuser.user_id))
        return {
            'form': form, 'dbuser': dbuser or self._DBUser,
            'user_name': user_name, 'action': action, 'profiles': profiles,
            'groups': groups, 'is_me': is_me, 'picture': picture,
            'tabset': Tabset(
                self._request, 'tabUser',
                self._DBUser.settings_tabs(self._request))}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='user_password_forgot',
        renderer='chrysalio:Templates/user_password_forgot.pt',
        permission=NO_PERMISSION_REQUIRED)
    def password_forgot(self):
        """Get email address to send a token to reset the password."""
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(
            String(), name='email',
            validator=All(Email(), Length(max=LABEL_LEN))))
        form = Form(self._request, schema)
        if form.validate():
            dbuser = self._request.dbsession.query(self._DBUser).filter(
                self._DBUser.email.ilike(form.values['email'])).first()
            if dbuser is None:
                self._request.session.flash(
                    _('Please use the email associated with your account.'),
                    'alert')
            elif self._send_reset_password_link(dbuser):  # pragma: nocover
                self._request.session.flash(
                    _('A link to reset your password has been sent by mail.'))
                log_info(
                    self._request, 'user_password_forgot', dbuser.login)
                return HTTPFound(self._request.route_path('home'))

        self._request.breadcrumbs(_('Forgot Password'), 1)
        return {'form': form}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='user_password_reset',
        renderer='chrysalio:Templates/user_password_reset.pt',
        permission=NO_PERMISSION_REQUIRED)
    def password_reset(self):
        """Reset the user password."""
        dbuser = self._get_user()
        token = '{0}{1}'.format(dbuser.login, date.today().isoformat())
        token = sha1(token.encode('utf8')).hexdigest()
        if token != self._request.matchdict['token']:
            raise HTTPNotFound(comment=_('This page does not exist anymore.'))

        password_min_length = self._request.registry['settings'][
            'password-min-length']
        schema = SchemaNode(Mapping())
        schema.add(SchemaNode(
            String(), name='password1',
            validator=All(
                Length(min=password_min_length),
                SameAs(self._request, 'password2'))))
        schema.add(SchemaNode(String(), name='password2'))
        form = Form(self._request, schema)
        if form.validate():
            dbuser.set_password(form.values['password1'])
            dbuser.password_mustchange = False
            self._request.session.flash(
                _('Your password has been changed. You can use it.'))
            log_info(self._request, 'user_password_reset')
            return HTTPFound(self._request.route_path('home'))

        self._request.breadcrumbs(_('Password Reset'), 1)
        return {'form': form}

    # -------------------------------------------------------------------------
    def can_view(self):
        """Check if the current user can view the user account.

        :rtype: bool
        """
        if self._request.has_permission('user-view'):
            return True
        user_id = self._request.matchdict.get('user_id')
        if user_id is None or (
                user_id.isdigit() and
                int(user_id) == self._request.session['user']['user_id']):
            return True
        return False

    # -------------------------------------------------------------------------
    def can_create(self):
        """Check if the current user can create a new user.

        :rtype: bool
        """
        if not self._request.has_permission('user-create'):
            return False
        return True

    # -------------------------------------------------------------------------
    def can_edit(self):
        """Check if the current user can edit an user account.

        :rtype: bool
        """
        if self._request.has_permission('user-edit'):
            return True
        user_id = self._request.matchdict['user_id']
        if user_id.isdigit() and \
           int(user_id) == self._request.session['user']['user_id']:
            return True
        return False

    # -------------------------------------------------------------------------
    def _get_user(self, my_user_id=''):
        """Return the SqlAlchemy object of the selected user or raise an
        HTTPNotFound exception.

        :param str my_user_id: (optional)
            User ID of current user.
        :rtype: .models.dbuser.DBUser
        """
        user_id = self._request.matchdict.get('user_id', my_user_id)
        if not user_id.isdigit():
            raise HTTPNotFound()
        dbuser = self._request.dbsession.query(self._DBUser).filter_by(
            user_id=int(user_id)).first()
        if dbuser is None:
            raise HTTPNotFound()
        return dbuser

    # -------------------------------------------------------------------------
    def _delete_users(self, user_ids):
        """Delete users.

        :param list user_ids:
            List of user IDs to delete.
        """
        user_ids = {int(k) for k in user_ids}

        # Do not delete myself
        if self._request.session['user']['user_id'] in user_ids:
            self._request.session.flash(
                _("You can't delete your own user."), 'alert')
            return

        # Do not delete an administrator
        if 1 in user_ids or \
           (not self._request.has_permission('system.administrator') and
            self._request.dbsession.query(self._DBUser.user_id).filter(
                self._DBUser.user_id.in_(user_ids)).filter_by(
                    status='administrator').first() is not None):
            self._request.session.flash(
                _("You can't delete an administrator."), 'alert')
            return

        # Delete
        deleted = []
        attachments = self._request.registry.settings.get('attachments')
        for dbuser in self._request.dbsession.query(self._DBUser).filter(
                self._DBUser.user_id.in_(user_ids)):
            if attachments and dbuser.attachments_key:
                attachment = join(
                    attachments, DBUser.attachments_dir,
                    dbuser.attachments_key)
                if exists(attachment):
                    rmtree(attachment)
            deleted.append(dbuser.login)
            self._request.dbsession.delete(dbuser)
        if deleted:
            log_info(self._request, 'user_delete', ' '.join(deleted))

    # -------------------------------------------------------------------------
    def _import_users(self):
        """Import users."""
        # Get current IDs
        user_logins = {
            k[0] for k in self._request.dbsession.query(DBUser.login)}

        # Update database
        web2db(self._request, self._xml2db[0], 'user')

        # Get new IDs
        user_logins = {k[0] for k in self._request.dbsession.query(
            DBUser.login)} - user_logins
        if user_logins:
            log_info(self._request, 'user_import', ' '.join(user_logins))

    # -------------------------------------------------------------------------
    def _users2response(self, user_ids):
        """Export users as an XML file embedded in a Pyramid response.

        :param list user_ids:
            List of user IDs to export.
        :rtype: :class:`pyramid.response.Response` or ``''``
        """
        dbitems = [
            k for k in self._request.dbsession.query(self._DBUser).filter(
                self._DBUser.user_id.in_(user_ids))
            if k.user_id != 1 or k.status != 'administrator']
        if not dbitems:
            self._request.session.flash(
                _('You cannot export the main administrator.'), 'alert')
            return ''
        filename = '{0}.{1}.xml'.format(
            len(dbitems) == 1 and dbitems[0].login or
            make_id(self._request.registry['settings']['title'], 'token'),
            self._DBUser.suffix)
        log_info(
            self._request, 'user_export',
            ' '.join([k.login for k in dbitems]))

        return db2web(self._request, dbitems, filename)

    # -------------------------------------------------------------------------
    def _send_invitation(self, user_ids):
        """Send an e-mail invitation to each user.

        :param list user_ids:
            List of user IDs to invite.
        :rtype: list
        :return:
            A list of errors.
        """
        site_title = self._request.registry['settings']['title']
        email_template = {
            'subject': _('Invitation to join ${n}', {'n': site_title}),
            'from': self._request.session['user']['email'],
            'html_template': self._invitation_html,
            'text_template': self._invitation_text}
        recipients = []
        for dbuser in self._request.dbsession.query(self._DBUser).filter(
                self._DBUser.user_id.in_(user_ids)):
            recipients.append({
                'to': dbuser.email,
                'site_title': site_title,
                'login': dbuser.login,
                'honorific': dbuser.honorific,
                'first_name': dbuser.first_name or '',
                'last_name': dbuser.last_name,
                'url': self._request.route_url('home')})
        log_info(
            self._request, 'user_send_invitation',
            ' '.join([k['login'] for k in recipients]))

        return Mailing(self._request).mailing(email_template, recipients)

    # -------------------------------------------------------------------------
    def _send_reset_password_link(self, dbuser):
        """Send an e-mail with a link to the page to reset the password.

        :type  dbuser: .models.dbuser.DBUser
        :param dbuser:
            The concerned user.
        :rtype: bool
        """
        site_title = self._request.registry['settings']['title']
        from_field = self._request.registry['settings']['email']
        email_template = {
            'subject': _('Password Reset on ${n}', {'n': site_title}),
            'from': from_field,
            'html_template': self._reset_password_html,
            'text_template': self._reset_password_text}

        token = '{0}{1}'.format(dbuser.login, date.today().isoformat())
        token = sha1(token.encode('utf8')).hexdigest()
        recipient = {
            'to': dbuser.email,
            'site_title': site_title,
            'login': dbuser.login,
            'honorific': dbuser.honorific,
            'first_name': dbuser.first_name or '',
            'last_name': dbuser.last_name,
            'url': self._request.route_url(
                'user_password_reset', user_id=dbuser.user_id, token=token)}

        error = Mailing(self._request).mailing(email_template, [recipient])
        if error:
            self._request.session.flash(error[0], 'alert')
            return False
        return True  # pragma: nocover

    # -------------------------------------------------------------------------
    def _save(self, dbuser, profiles, groups, values):
        """Save a user settings.

        :type  dbuser: :class:`.models.dbuser.DBUser` or ``None``
        :param dbuser:
            User to save.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :param dict groups:
            A dictionary such as ``{group_id: (label, description),...}``.
        :param dict values:
            Form values.
        :rtype: :class:`~.models.dbuser.DBUser` instance or ``None``
        """
        creation = dbuser is None
        dbuser = dbuser or self._DBUser()

        # Update user
        record = {k: values[k] for k in values if hasattr(self._DBUser, k)}
        record['password'] = values['password1'] or dbuser.password
        error = dbuser.record_format(record)
        if error:  # pragma: nocover
            self._request.session.flash(error, 'alert')
            return None
        modified = creation
        record.update({
            k:  None for k in values
            if not values[k] and hasattr(self._DBUser, k)})
        for field in record:
            if getattr(dbuser, field) != record[field]:
                modified = True
                setattr(dbuser, field, record[field])
        if not creation and values['password1']:
            modified = True
            dbuser.password_update = datetime.now()

        # Save
        if creation:
            try:
                self._request.dbsession.add(dbuser)
                self._request.dbsession.flush()
            except (IntegrityError, FlushError):
                self._request.session.flash(
                    _('This user already exists.'), 'alert')
                return None

        # Update extra informations
        modified |= self._save_extra(dbuser, profiles, groups, values)

        if modified:
            dbuser.account_update = datetime.now()

        return dbuser

    # -------------------------------------------------------------------------
    def _save_extra(self, dbuser, profiles, groups, values):
        """Save extra information on a user .

        :type  dbuser: :class:`.models.dbuser.DBUser` or ``None``
        :param dbuser:
            User to save.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :param dict groups:
            A dictionary such as ``{group_id: (label, description),...}``.
        :param dict values:
            Form values.
        :rtype: bool
        """
        modified = False
        if not self._request.has_permission('user-create'):
            return False

        # Update profiles
        user_profiles = {k.profile_id: k for k in dbuser.profiles}
        for profile_id in profiles:
            value = values['pfl:{0}'.format(profile_id)]
            if value and profile_id not in user_profiles:
                modified = True
                self._request.dbsession.add(DBUserProfile(
                    user_id=dbuser.user_id, profile_id=profile_id))
            elif not value and profile_id in user_profiles:
                modified = True
                self._request.dbsession.delete(
                    self._request.dbsession.query(DBUserProfile).filter_by(
                        user_id=dbuser.user_id, profile_id=profile_id)
                    .first())

        # Update groups
        user_groups = {k.group_id: k for k in dbuser.groups}
        for group_id in groups:
            value = values['grp:{0}'.format(group_id)]
            if value and group_id not in user_groups:
                modified = True
                self._request.dbsession.add(DBGroupUser(
                    group_id=group_id, user_id=dbuser.user_id))
            elif not value and group_id in user_groups:
                modified = True
                self._request.dbsession.delete(
                    self._request.dbsession.query(DBGroupUser).filter_by(
                        group_id=group_id, user_id=dbuser.user_id)
                    .first())

        return modified



"""Pyramid session functionality."""

from pyramid.session import SignedCookieSessionFactory
from pyramid.session import JSONSerializer, PickleSerializer
from pyramid.config import Configurator


# =============================================================================
def includeme(configurator):
    """Function to include session from Pyramid.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    if isinstance(configurator, Configurator):
        configurator.set_session_factory(
            SignedCookieSessionFactory(
                configurator.get_settings().get(
                    'site.uid', 'chrysalio_sekret'),
                serializer=JSONSerializerWithPickleFallback()))


# =============================================================================
class JSONSerializerWithPickleFallback(object):
    """To support upcoming changes to ISession in Pyramid 2.0."""

    # -------------------------------------------------------------------------
    def __init__(self):
        """Constructor method."""
        self.json = JSONSerializer()
        self.pickle = PickleSerializer()

    # -------------------------------------------------------------------------
    def dumps(self, value):
        """Serialize obj to a JSON formatted str."""
        return self.json.dumps(value)

    # -------------------------------------------------------------------------
    def loads(self, value):
        """Deserialize s (a str, bytes or bytearray instance containing a JSON
        document) to a Python object.
        """
        try:
            return self.json.loads(value)
        except ValueError:  # pragma: nocover
            return self.pickle.loads(value)
